/****************************************************************************
 * LAPACK, BLAS AND ATLAS functions definitions used in Vector, Matrix,     *
 * SquareMatrix and Tensor class.                                           *
 *                                                                          *
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#ifndef _FUNCTIONS_H_
#define _FUNCTIONS_H_

extern "C"
{
  // include atlas functions
#include <cblas.h>
#include <clapack.h>

  extern void dgesv_(const int *N, const int *NRHS, double *A,
		     const int *LDA, int *IPIV, double *B, const int *LDB,
		     int *INFO) ;
  //  Purpose
  //  =======
  //
  //  DGESV computes the solution to a real system of linear equations
  //     A * X = B,
  //  where A is an N-by-N matrix and X and B are N-by-NRHS matrices.
  //
  //  The LU decomposition with partial pivoting and row interchanges is
  //  used to factor A as
  //     A = P * L * U,
  //  where P is a permutation matrix, L is unit lower triangular, and U is
  //  upper triangular.  The factored form of A is then used to solve the
  //  system of equations A * X = B.
  //
  //  Arguments
  //  =========
  //
  //  N       (input) INTEGER
  //          The number of linear equations, i.e., the order of the
  //          matrix A.  N >= 0.
  //
  //  NRHS    (input) INTEGER
  //          The number of right hand sides, i.e., the number of columns
  //          of the matrix B.  NRHS >= 0.
  //
  //  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
  //          On entry, the N-by-N coefficient matrix A.
  //          On exit, the factors L and U from the factorization
  //          A = P*L*U; the unit diagonal elements of L are not stored.
  //
  //  LDA     (input) INTEGER
  //          The leading dimension of the array A.  LDA >= max(1,N).
  //
  //  IPIV    (output) INTEGER array, dimension (N)
  //          The pivot indices that define the permutation matrix P;
  //          row i of the matrix was interchanged with row IPIV(i).
  //
  //  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
  //          On entry, the N-by-NRHS matrix of right hand side matrix B.
  //          On exit, if INFO = 0, the N-by-NRHS solution matrix X.
  //
  //  LDB     (input) INTEGER
  //          The leading dimension of the array B.  LDB >= max(1,N).
  //
  //  INFO    (output) INTEGER
  //          = 0:  successful exit
  //          < 0:  if INFO = -i, the i-th argument had an illegal value
  //          > 0:  if INFO = i, U(i,i) is exactly zero.  The factorization
  //                has been completed, but the factor U is exactly
  //                singular, so the solution could not be computed.
  //
  //  =====================================================================


  extern void dspsv_( const char *UPLO, const int *N, const int *NRHS, 
		      double *AP, const int *IPIV, double *B, const int *LDB, 
		      int *INFO ) ;
  //  Purpose
  //  =======
  //
  //  DSPSV computes the solution to a real system of linear equations
  //     A * X = B,
  //  where A is an N-by-N symmetric matrix stored in packed format and X
  //  and B are N-by-NRHS matrices.
  //
  //  The diagonal pivoting method is used to factor A as
  //     A = U * D * U**T,  if UPLO = 'U', or
  //     A = L * D * L**T,  if UPLO = 'L',
  //  where U (or L) is a product of permutation and unit upper (lower)
  //  triangular matrices, D is symmetric and block diagonal with 1-by-1
  //  and 2-by-2 diagonal blocks.  The factored form of A is then used to
  //  solve the system of equations A * X = B.
  //
  //  Arguments
  //  =========
  //
  //  UPLO    (input) CHARACTER
  //          = 'U':  Upper triangle of A is stored;
  //          = 'L':  Lower triangle of A is stored.
  //
  //  N       (input) INTEGER
  //          The number of linear equations, i.e., the order of the
  //          matrix A.  N >= 0.
  //
  //  NRHS    (input) INTEGER
  //          The number of right hand sides, i.e., the number of columns
  //          of the matrix B.  NRHS >= 0.
  //
  //  AP      (input/output) DOUBLE PRECISION array, dimension (N*(N+1)/2)
  //          On entry, the upper or lower triangle of the symmetric matrix
  //          A, packed columnwise in a linear array.  The j-th column of A
  //          is stored in the array AP as follows:
  //          if UPLO = 'U', AP(i + (j-1)*j/2) = A(i,j) for i<=j;
  //          if UPLO = 'L', AP(i + (j-1)*(2n-j)/2) = A(i,j) for j<=i<=n.
  //          See below for further details.
  //
  //          On exit, the block diagonal matrix D and the multipliers used
  //          to obtain the factor U or L from the factorization
  //          A = U*D*U**T or A = L*D*L**T as computed by DSPTRF, stored as
  //          a packed triangular matrix in the same storage format as A.
  //
  //  IPIV    (output) INTEGER array, dimension (N)
  //          Details of the interchanges and the block structure of D, as
  //          determined by DSPTRF.  If IPIV(k) > 0, then rows and columns
  //          k and IPIV(k) were interchanged, and D(k,k) is a 1-by-1
  //          diagonal block.  If UPLO = 'U' and IPIV(k) = IPIV(k-1) < 0,
  //          then rows and columns k-1 and -IPIV(k) were interchanged and
  //          D(k-1:k,k-1:k) is a 2-by-2 diagonal block.  If UPLO = 'L' and
  //          IPIV(k) = IPIV(k+1) < 0, then rows and columns k+1 and
  //          -IPIV(k) were interchanged and D(k:k+1,k:k+1) is a 2-by-2
  //          diagonal block.
  //
  //  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
  //          On entry, the N-by-NRHS right hand side matrix B.
  //          On exit, if INFO = 0, the N-by-NRHS solution matrix X.
  //
  //  LDB     (input) INTEGER
  //          The leading dimension of the array B.  LDB >= max(1,N).
  //
  //  INFO    (output) INTEGER
  //          = 0:  successful exit
  //          < 0:  if INFO = -i, the i-th argument had an illegal value
  //          > 0:  if INFO = i, D(i,i) is exactly zero.  The factorization
  //                has been completed, but the block diagonal matrix D is
  //                exactly singular, so the solution could not be
  //                computed.
  //
  //  Further Details
  //  ===============
  //
  //  The packed storage scheme is illustrated by the following example
  //  when N = 4, UPLO = 'U':
  //
  //  Two-dimensional storage of the symmetric matrix A:
  //
  //     a11 a12 a13 a14
  //         a22 a23 a24
  //             a33 a34     (aij = aji)
  //                 a44
  //
  //  Packed storage of the upper triangle of A:
  //
  //  AP = [ a11, a12, a22, a13, a23, a33, a14, a24, a34, a44 ]
  //
  //  =====================================================================

  extern void dsyev_(const char *JOBZ, const char *UPLO, const int *N, double *A,
		     const int *LDA, double *W, double *WORK, const int *LWORK, 
		     int *INFO) ;
  //  Purpose
  //  =======
  //
  //  DSYEV computes all eigenvalues and, optionally, eigenvectors of a
  //  real symmetric matrix A.
  //
  //  Arguments
  //  =========
  //
  //  JOBZ    (input) CHARACTER
  //          = 'N':  Compute eigenvalues only;
  //          = 'V':  Compute eigenvalues and eigenvectors.
  //
  //  UPLO    (input) CHARACTER
  //          = 'U':  Upper triangle of A is stored;
  //          = 'L':  Lower triangle of A is stored.
  //
  //  N       (input) INTEGER
  //          The order of the matrix A.  N >= 0.
  //
  //  A       (input/output) DOUBLE PRECISION array, dimension (LDA, N)
  //          On entry, the symmetric matrix A.  If UPLO = 'U', the
  //          leading N-by-N upper triangular part of A contains the
  //          upper triangular part of the matrix A.  If UPLO = 'L',
  //          the leading N-by-N lower triangular part of A contains
  //          the lower triangular part of the matrix A.
  //          On exit, if JOBZ = 'V', then if INFO = 0, A contains the
  //          orthonormal eigenvectors of the matrix A.
  //          If JOBZ = 'N', then on exit the lower triangle (if UPLO='L')
  //          or the upper triangle (if UPLO='U') of A, including the
  //          diagonal, is destroyed.
  //
  //  LDA     (input) INTEGER
  //          The leading dimension of the array A.  LDA >= max(1,N).
  //
  //  W       (output) DOUBLE PRECISION array, dimension (N)
  //          If INFO = 0, the eigenvalues in ascending order.
  //
  //  WORK    (workspace/output) DOUBLE PRECISION array, dimension (MAX(1,LWORK))
  //          On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
  //
  //  LWORK   (input) INTEGER
  //          The length of the array WORK.  LWORK >= max(1,3*N-1).
  //          For optimal efficiency, LWORK >= (NB+2)*N,
  //          where NB is the blocksize for DSYTRD returned by ILAENV.
  //
  //          If LWORK = -1, then a workspace query is assumed; the routine
  //          only calculates the optimal size of the WORK array, returns
  //          this value as the first entry of the WORK array, and no error
  //          message related to LWORK is issued by XERBLA.
  //
  //  INFO    (output) INTEGER
  //          = 0:  successful exit
  //          < 0:  if INFO = -i, the i-th argument had an illegal value
  //          > 0:  if INFO = i, the algorithm failed to converge; i
  //                off-diagonal elements of an intermediate tridiagonal
  //                form did not converge to zero.

  extern void dgelss_(const int *M, const int *N, const int *NRHS, double *A, const int *LDA,
		      double *B, const int *LDB, double *S, const double *RCOND, int *RANK,
		      double *WORK, int *LWORK, int *INFO) ;
//  Purpose
//  =======
//
//  DGELSS computes the minimum norm solution to a real linear least
//  squares problem:
//
//  Minimize 2-norm(| b - A*x |).
//
//  using the singular value decomposition (SVD) of A. A is an M-by-N
//  matrix which may be rank-deficient.
//
//  Several right hand side vectors b and solution vectors x can be
//  handled in a single call; they are stored as the columns of the
//  M-by-NRHS right hand side matrix B and the N-by-NRHS solution matrix
//  X.
//
//  The effective rank of A is determined by treating as zero those
//  singular values which are less than RCOND times the largest singular
//  value.
//
//  Arguments
//  =========
//
//  M       (input) INTEGER
//          The number of rows of the matrix A. M >= 0.
//
//  N       (input) INTEGER
//          The number of columns of the matrix A. N >= 0.
//
//  NRHS    (input) INTEGER
//          The number of right hand sides, i.e., the number of columns
//          of the matrices B and X. NRHS >= 0.
//
//  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
//          On entry, the M-by-N matrix A.
//          On exit, the first min(m,n) rows of A are overwritten with
//          its right singular vectors, stored rowwise.
//
//  LDA     (input) INTEGER
//          The leading dimension of the array A.  LDA >= max(1,M).
//
//  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
//          On entry, the M-by-NRHS right hand side matrix B.
//          On exit, B is overwritten by the N-by-NRHS solution
//          matrix X.  If m >= n and RANK = n, the residual
//          sum-of-squares for the solution in the i-th column is given
//          by the sum of squares of elements n+1:m in that column.
//
//  LDB     (input) INTEGER
//          The leading dimension of the array B. LDB >= max(1,max(M,N)).
//
//  S       (output) DOUBLE PRECISION array, dimension (min(M,N))
//          The singular values of A in decreasing order.
//          The condition number of A in the 2-norm = S(1)/S(min(m,n)).
//
//  RCOND   (input) DOUBLE PRECISION
//          RCOND is used to determine the effective rank of A.
//          Singular values S(i) <= RCOND*S(1) are treated as zero.
//          If RCOND < 0, machine precision is used instead.
//
//  RANK    (output) INTEGER
//          The effective rank of A, i.e., the number of singular values
//          which are greater than RCOND*S(1).
//
//  WORK    (workspace/output) DOUBLE PRECISION array, dimension (MAX(1,LWORK))
//          On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
//
//  LWORK   (input) INTEGER
//          The dimension of the array WORK. LWORK >= 1, and also:
//          LWORK >= 3*min(M,N) + max( 2*min(M,N), max(M,N), NRHS )
//          For good performance, LWORK should generally be larger.
//
//          If LWORK = -1, then a workspace query is assumed; the routine
//          only calculates the optimal size of the WORK array, returns
//          this value as the first entry of the WORK array, and no error
//          message related to LWORK is issued by XERBLA.
//
//  INFO    (output) INTEGER
//          = 0:  successful exit
//          < 0:  if INFO = -i, the i-th argument had an illegal value.
//          > 0:  the algorithm for computing the SVD failed to converge;
//                if INFO = i, i off-diagonal elements of an intermediate
//                bidiagonal form did not converge to zero.
}

#endif
