/****************************************************************************
 * Matrix class for manipulating two-dimesional arrays in numerical         *
 * computations. This class uses a general matrix storage                   *
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <iostream>
using std::ostream ;
using std::istream ;

#include "Functions.H" 
#include "Const_num.H"

class Vector ;

#ifndef _STATUS_
#define _STATUS_

enum Status { NORMAL, LU_DECOMP, QR_DECOMP, SVD_DECOMP,
	      CHOLESKY_DECOMP, BALANCED, UNINITIALIZED } ;

#endif

class Matrix
{
  friend ostream &operator<<( ostream &, const Matrix & ) ;
  friend istream &operator>>( istream &, Matrix & ) ;
public:
  Matrix() ;                  // default constructor
  Matrix(int,int) ;     // initialize all elements to zero 
  Matrix(const Matrix &) ;    // copy constructor
  virtual ~Matrix() ;         // destructor

  // test if matrix is uninitialized.
  // i.e. _rows and _cols has not been initialized
  bool IsNull() const ;   // return true if it is null false otherwise
 
  int Rows() const ;   // return numbers of rows
  int Cols() const ;   // return numbers of cols  

  // this function return a const pointer to the matrix data.
  // It is mainly intended to be used inside Vector class. 
  const double *Data() const ; 

  // this function return a pointer to the matrix data.
  // It is mainly intended to be used inside SquareMatrix class.
  // Note that this function could be used as a modifiable lvalue and
  // that, if is use in a wrong way, can cause problems 
  double *Data() ; 
  
  // this function set the status of the matrix; encoding using the
  // variable Status defined above. Use it carefully and just if
  // you now what you are doing.
  void SetStatus(Status) ; 

  // this function return the status of the matrix; encoding using the
  // variable Status defined above
  Status GetStatus() const ; 
 
  // This function clear a matrix, that is, transforming it to a non
  // initialized matrix. It has the same effect that the destructor. 
  void Clear() ;

  // this function set the number of rows and columns in matrix m.
  // The matrix must be uninitialized.  Set all elements to zero.
  // i.e. Matrix m ; m.SetSize(2,2) ;
  void SetSize(int row, int col) ;

  const Matrix &operator=( const Matrix & ) ;  // assigment operator
 
  // subscript operator for non-const objects returns modifiable lvalue
  double &operator()( int , int ) ;

  // subscript operator for const objects returns rvalue
  double operator()( int , int ) const ;

  // set all the elements of matrix m to the value x
  // m(i,j) = x
  void SetAll( double ) ;

  // sets all the elements to the value zero
  // m(i,j) = 0
  void SetZero() ;

  // multiplies the elements of matrix a by the constant factor x.
  // a'(i,j) = x * a(i,j)
  void Scale( double ) ;

  // adds the constant value x to the elements of the matrix a.
  // a'(i,j) = a(i,j) + x
  void AddConstant( double ) ;

  // += overloded operator for non-const objects.
  // a'(i,j) = a(i,j) + b(i,j)
  Matrix &operator+=( const Matrix & ) ;

  // -= overloded operator for non-const objects.
  // a'(i,j) = a(i,j) - b(i,j)
  Matrix &operator-=( const Matrix & ) ;

  // *= overloded operator for non-const objects.
  // a'(i,j) = a(i,j) * b(i,j)
  Matrix &operator*=( const Matrix & ) ;

  // this function returns the maximum value in the matrix.
  double MaxValue() const ;

  // this function returns the minimum value in the matrix..
  double MinValue() const ;

  // this function returns the minimim and the maximum values in
  // the matrix m, storing in min_out and max_out.
  void MinMaxValue( double &min_out, double &max_out) const ;

  // this function returns the index of the maximum value in the matrix m.
  // when there are several maximum elements then the lowest index is
  // returned
  void MaxIndex( int &imax , int &jmax ) const ;

  // this function returns the index of the minimum value in the matrix m.
  // when there are several minimum elements then the lowest index is
  // returned
  void MinIndex( int &imin , int &jmin ) const ;

  // this function returns the index of the maximum and minimum values
  // in the matrix m, storing them in imin and imax. when there are several 
  // maximum or minimum elements then the lowest indeces are returned
  void MinMaxIndex( int &imin, int &jmin , 
		    int &imax, int &jmax ) const ;

  // this function copies the elements of the i-th row of the matrix m
  // into the vector v.
  void GetRow(int , Vector &) const ;

  // this function copies the elements of the j-th column of the matrix m
  // into the vector v.
  void GetCol(int, Vector &) const ;

  // this function copies the elements of the vector v into the i-th row 
  // of the matrix m.
  void SetRow(const Vector & , int) ;

  // this function copies the elements of the vector v into the j-th column 
  // of the matrix m.
  void SetCol(const Vector & , int) ;

  // this function makes the matrix dest the transpose of the matrix src by
  // copying the elements of src into dest. This function works for all
  // matrices provided that the matrix dest is uninitialized.
  // i.e. Matrix dest ; src.Transpose(dest) ;
  void Transpose(Matrix &dest) const ;

  // this function compute the matrix-matrix product and sum C' = A B + C
  // i.e. Matrix A(2,2) ; A.SetAll(1.0) ; Matrix B(2,2) ;
  //      B.setAll(2.0) ; Matrix C(2,2) ; C.SetAll(3.0) ; C.dgemm(A,B) ; 
  void dgemm(const Matrix &A, const Matrix &B) const ;

  // Computes the minimum norm solution to a real linear least
  // squares problem:
  // Minimize 2-norm(| b - A*x |).
  // using the singular value decomposition (SVD) of A. A is an M-by-N
  // matrix which may be rank-deficient.
  // The condition number of A in the 2-norm = S(1)/S(min(m,n)).
  // where S are the singular values of A.
  void SVD_Solve(Vector &x,double &cond_number) ;

private:
  int _rows ;                  // number of rows
  int _cols ;                  // number of columns
  Status _status ;             // store the status of the general matrix.
                               // i.e. after an LU decomposition has been modified
  double *_data ;              // pointer to first element of pointer-based array
                               // column major order.
} ; // end class Matrix

#endif
