/****************************************************************************
 * SparseMatrix class for manipulating two-dimesional arrays in numerical   *
 * computations. This class uses the Harwell-Boing format.                  *
 * Arrays are assumed general square matrices.                              *
 * Copyright (C) 2010 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 * 09/30/10    Mario Juha   Added the overloading >> and << operators.      *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#ifndef _SPARSEMATRIX_H_
#define _SPARSEMATRIX_H_

#include <iostream>
using std::ostream ;
using std::istream ;

#include "slu_ddefs.h"       // SuperLU function prototypes

class Vector ;

#ifndef _STATUS_
#define _STATUS_

enum Status { NORMAL, LU_DECOMP, QR_DECOMP, SVD_DECOMP,
	      CHOLESKY_DECOMP, BALANCED, UNINITIALIZED } ;

#endif


class SparseMatrix
{
  // this operator output the sparse matrix using the following
  // format:
  // first line  : _size  _nnz
  // second line : _rowind
  // third line  : _colptr
  // fourth line : _nzval
  friend ostream &operator<<( ostream &, const SparseMatrix & ) ;
  // this operator read in the sparse matrix using the following
  // format:
  // first line  : _size  _nnz
  // second line : _rowind
  // third line  : _colptr
  // fourth line : _nzval
  //
  // Note: The sparse matrix to fill in must be uninitialized.
  // ====
  friend istream &operator>>( istream &, SparseMatrix & ) ;
public:
  SparseMatrix() ;                  // default constructor
  // Parametrized constructor 
  SparseMatrix( int size, int nnz) ;
  // Parametrized constructor. It will free the arrays passed to the
  // constructor when the destructor is called. Use it carefully!!!
  SparseMatrix( int size, int nnz, double *a, int *asub, int *xa) ;
  SparseMatrix(const SparseMatrix &A) ;    // copy constructor
  virtual ~SparseMatrix() ;         // destructor

  // test if matrix is uninitialized.
  // i.e. _rows and _cols has not been initialized
  bool IsNull() const ;   // return true if it is null false otherwise
 
  int Size() const ;   // return numbers of rows or columns

  int NumberNonZeros() const ;  // return number of non zero values 

  // sets all the elements to the value zero
  // m(i,j) = 0
  void SetZero() ;

  // this function set the nonzero coefficients of matrix A into array
  // of _nzval.
  void SetNZValues(int index, double val) ;
  
  // this function set the row indices
  void SetRowIndex(int index, int val ) ;

  // this function set the beginning of each column in the coefficient
  // and row index array.
  void SetColIndex(int index, int val) ;

  // this function return a const pointer to the row index data.
  const int *GetRowIndexArray() const ; 

  // this function return a pointer to the row index data.
  // Note that this function could be used as a modifiable lvalue and
  // that, if is use in a wrong way, can cause problems 
  int *GetRowIndexArray() ; 

  // this function return a const pointer to the col index data.
  const int *GetColIndexArray() const ; 
  
  // this function return a pointer to the col index data.
  // Note that this function could be used as a modifiable lvalue and
  // that, if is use in a wrong way, can cause problems 
  int *GetColIndexArray() ; 

  // this function return the status of the sparse matrix; encoding using the
  // variable Status defined above
  Status GetStatus() const ; 
  
  // This function clear a matrix, that is, transforming it to a non
  // initialized matrix. It has the same effect that the destructor. 
  void Clear() ;

  // this function set the size in matrix A.
  // The matrix must be uninitialized.  Set all elements to zero.
  // i.e. SparseMatrix A ; A.SetSize(4,2) ;
  void SetSize(int size, int nnz) ;
  
  // set all the elements of sparse matrix m to the value x
  // m(i,j) = x
  void SetAll( double x ) ;
  
  // multiplies the elements of sparse matrix a by the constant factor x.
  // a'(i,j) = x * a(i,j)
  void Scale( double x ) ;

  // subscript operator for non-const objects returns modifiable lvalue
  double &operator()( int row, int col ) ;
  
  // subscript operator for const objects returns rvalue
  double operator()( int row , int col ) const ;

  // This function computes the solution to a real system of linear
  // equations A * x = b, using SuperLU library.  where A is an N-by-N
  // sparse (Harwell-Boing format) matrix and x and b are N-by-1
  // vectors.
  //
  //  The LU decomposition with partial pivoting and row interchanges is
  //  used to factor A as
  //     A = P * L * U,
  //  where P is a permutation matrix, L is unit lower triangular, and U is
  //  upper triangular.  The factored form of A is then used to solve the
  //  system of equations A * x = b.
  void LinSolve(Vector &x) const ;

  // This function computes the solution to a real system of linear
  // equations A * x = b, using SuperLU library.  where A is an N-by-N
  // sparse (Harwell-Boing format) matrix and x and b are N-by-1
  // vectors.
  //
  //  The LU decomposition with partial pivoting and row interchanges is
  //  used to factor A as
  //     A = P * L * U,
  //  where P is a permutation matrix, L is unit lower triangular, and
  //  U is upper triangular.  The factored form of A is then used to
  //  solve the system of equations A * x = b. An expert driver
  //  algorithm is used and the inverse of the condition number
  //  (rcond) and the reciprocal pivot growth factor (rpg) are
  //  returned. This is generally more expensive that LinSolve.
  void LinSolveExpert(Vector &x, double &rcond, double &rpg) const ;
    
  // This function computes the solution to a real system of linear
  // equations A * x = b, using the conjugate gradient method.  where A
  // is an N-by-N symmetric and positive definite sparse (Harwell-Boing
  // format) matrix and x and b are N-by-1 vectors. Recommended over
  // LinSolve is A is extremely large. This function implement the
  // algorithm presented in:
  //
  // D.Kincaid and W.Cheney. Numerical Analysis: Mathematics of Scientifics
  // Computing. Third Edition. Brooks/Cole Series in Advanced Mathematics.
  // USA,2002. pag. 238
  //
  // Note: DO NOT USE if A is not symmetric and positive definite.
  // ----
  void ConjGradSolve(Vector &x, const int MaxIter) const ;

  // this function compute the matrix-vector product and sum y' = A x + y
  // where A is a sparse matrix and x and y are vectors.
  // i.e. A.dgemv(x,y) ;
  void dgemv(const Vector &x, Vector &y) const ;
  
  // this function compute the matrix-matrix product and sum C' = A B + C
  // where A is a sparse matrix and B and C are general matrices.
  // i.e. A.dgemm(B,C) ;
  void dgemm(const Matrix &B, Matrix &C) const ;
  
private:
  int _size ;                  // number of rows or columns
  int _nnz ;                   // number of nonzeros in the matrix
  Status _status ;             // store the status of the sparse matrix
                               // i.e. after an LU decomposition has been modified
  double *_nzval ;             // array of nonzero values packed by column
  int *_rowind ;               // array of row indices of the nonzeros
  int *_colptr ;               // _colptr[j] stores the location in nzval[] and
                               // _rowind[] which starts column j. It has ncol+1
                               // entries, and colptr[ncol] = nnz. 
} ; // end class SparseMatrix

#endif
