/****************************************************************************
 * SquareMatrix class for manipulating two-dimesional arrays in numerical   *
 * computations. This class uses a general matrix storage                   *
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#ifndef _SQUAREMATRIX_H_
#define _SQUAREMATRIX_H_

#include <iostream>
using std::ostream ;
using std::istream ;

#include "Functions.H" 
#include "Const_num.H"


class SquareMatrix : public Matrix
{
public:
  SquareMatrix() ;                  // default constructor
  SquareMatrix(int) ;     // initialize all elements to zero 
  SquareMatrix(const SquareMatrix &) ;    // copy constructor
  virtual ~SquareMatrix() ;                 // destructor

  // this function set the number of rows and columns in matrix m.
  // The matrix must be uninitialized.  Set all elements to zero.
  // i.e. SquareMatrix m ; m.SetSize(2) ;
  virtual void SetSize(int) ;

  // this function sets the elements of the matrix m to the corresponding
  // elements of the identity matrix, m(i,j) = \delta(i,j)
  void SetIdentity() ;

  // this function factorize the square matrix A of M x N into the LU decomposition
  // PA = LU. 
  // i.e:
  // SquareMatrix sqm(2) ; 
  // sqm(0,0) =0.18; sqm(0,1) = 0.60;
  // sqm(1,0) =0.41; sqm(1,1) = 0.24;
  // sqm.LU_Factor() ;
  void LU_Factor() ;

  //  this function solves a system of linear equations
  //     A * x = b
  //  with a general N-by-N matrix A using the LU factorization computed
  //  by LU_Factor().
  void LU_Solve(Vector &x) ;
  
  // This function computes the solution to a real system of linear equations
  //     A * x = b, using ATLAS library.
  //  where A is an N-by-N matrix and x and b are N-by-1 vectors.
  //
  //  The LU decomposition with partial pivoting and row interchanges is
  //  used to factor A as
  //     A = P * L * U,
  //  where P is a permutation matrix, L is unit lower triangular, and U is
  //  upper triangular.  The factored form of A is then used to solve the
  //  system of equations A * x = b.
  void LinSolve(Vector &x) ;  

  // this funtion solve the system A x = b in-place using the Cholesky
  // decomposition of A. On input x should contain the right-hand side b,
  // which is replaced by the solution on output.
  // The Cholesky decomposition of A can only be carried out when all
  // the eigenvalues of the matrix are positive. If the matrix is not
  // symmetric positive-definite then the decomposition will fail.
  void Cholesky_Solve(Vector &x) ;

  // this function compute the inverse of a matrix A from its LU
  // decomposition (LU,p), storing the results in the matrix
  // inverse. The inverse is computed by solving the system A x = b
  // for each column of the identity matrix. The inverse matrix
  // passed must be uninitialized.
  void Invert(SquareMatrix &inverse) const ;

  // this function compute the determinant of a matrix A from its LU
  // decomposition, LU. The determinant is computed as the product of
  // the diagonal elements of U and the sign of the row permutation.
  // Note: This function does not check for underflow or overflow.
  double Det() const ;

} ; // end class SquareMatrix


/****************************************************************************
 * Tensor class for manipulating second order tensors in numerical          *
 * computations.                                                            *
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 04/10/10    Mario Juha     Original code                                 *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

class Tensor : public SquareMatrix
{
public:
  Tensor() ;                  // default constructor
  Tensor(int) ;            // initialize all elements to zero 
  Tensor(const Tensor &) ;    // copy constructor
  virtual ~Tensor() ;    

  //-----------------------
  // this function set the number of rows and columns in tensor A.
  // The tensor must be uninitialized. Acceptable size are 2 or 3.
  // Initialized all the elements to zero.
  // i.e. Tensor A ; A.SetSize(3) ;
  virtual void SetSize(int size) ;

  // this function calculate the invariants of a REAL SYMMETRIC second
  // order tensor A, solving first for the eigenvalues and after that,
  // using the following formulas:
  // 2D:
  //         invariants(0) = a1 + a2
  //         invariants(1) = a1*a2 
  // 3D: 
  //         invariants(0) = a1 + a2 + a3 
  //         invariants(1) = a1*a2 + a1*a3 + a2*a3
  //         invariants(2) = a1*a2*a3 
  // Note that a second order tensor is represented by a square matrix
  // of size 3 x 3 in 3D space and a matrix of 2 x 2 in a 2D space. If
  // you try to use this function for a matrix with a different size
  // the program will abort.  The result will be unexpected if you try
  // to use it with an non-symmetric tensor.
  void Invariants(Vector &invariants) const ;

  // this function compute the trace of a tensor. tr(A) = A_ii.
  double Trace() const ;

  // this function compute the contraction of the tensor A and
  // B. Using A : B = A_ij B_ij.
  double Contraction(const Tensor &B) const ;

  // this function compute the symmetric and skew part of a tensor.
  // A = S + W. S = 1/2 * (A + A^T), W = 1/2 * (A - A^T)
  // S = Symmetric part of A
  // W = Skew part of A
  void SymmSkew(Tensor &S, Tensor &W) const ; 

  // this function decompose a tensor A in its so-called spherical
  // part and its deviatoric part. i.e.
  // A_ij = alpha*delta_ij + dev A_ij
  // A = S + Dev
  // S = spherical part
  // Dev =  deviator part
  void SphericalDeviator(Tensor &S, Tensor &Dev) const ;

  // this function compute the polar decomposition of the deformation
  // gradient F. That is, F = RU. This function just have senses for the
  // polar decomposition of the deformation gradient. 
  // F = deformation gradient
  // R = orthogonal rotation matrix
  // U = symmetric strecth matrix
  // See: K.Bathe. Finite Element Procedures. Prentice Hall. NJ, 1996.
  // p. 508 - 510.
  void PolarDecomposition(Tensor &R, Tensor &U) const ;

  // this function evaluate the transformation between the Cauchy
  // stress tensor and the Second Piola-Kirchhoff stress tensor. S is
  // the second piola-kirchhoff stress tensor and should be
  // initialized.
  void CauchyPK2(Tensor &S, const Tensor &F) const ;

  // this function evaluate the transformation between the Second
  // Piola-Kirchhoff stress tensor and the Cauche stress tensor. sigma
  // is the Cauchy stress tensor and should be initialized.
  void PK2Cauchy(Tensor &sigma, const Tensor &F) const ;

  // this function evaluate the transformation between the second
  // Piola-Kirchhoff stress tensor and the nominal stress tensor. P is
  // the nominal stress tensor and should be initialized.
  void PK2Nominal(Tensor &P, const Tensor &F) const ;

  // this function evaluate the transformation between the
  // Green-Lagrange strain tensor and the Euler-Almansi strain tensor.
  // The Euler-Almansi strain tensor should be initialized.
  void GreenAlmansi(Tensor &almansi, const Tensor &F) const ;

  // this function transform a vector into a tensor following the
  // kinetic Voigt rule.  
  // See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
  // Continua and Structures. Wiley. England. 2000. p.615-616.
  // Note the tensor should be uninitialized.
  void KineticVoigtRule(const Vector &) ;

  // this function transform a vector into a tensor following the
  // kinematic Voigt rule.  
  // See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
  // Continua and Structures. Wiley. England. 2000. p.615-616.
  // Note the tensor should be uninitialized.
  void KinematicVoigtRule(const Vector &) ;

  // this function transform a tensor into a vector following the
  // kinetic Voigt rule.  
  // See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
  // Continua and Structures. Wiley. England. 2000. p.615-616.
  // Note the vector to pass should be uninitialized.
  void KineticVoigtRule_TV(Vector &) const ;

  // this function transform a tensor into a vector following the
  // kinematic Voigt rule.  
  // See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
  // Continua and Structures. Wiley. England. 2000. p.615-616.
  // Note the vector to pass should be uninitialized.
  void KinematicVoigtRule_TV(Vector &) const ;

} ;

#endif
