/****************************************************************************
 * Vector class for manipulating one-dimesional arrays in numerical         *
 * computations.                                                            *
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/10/10    Mario Juha     Original code                                 *
 * 11/26/10    Mario Juha   Adeed function Clear()                          *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#ifndef _VECTOR_H_
#define _VECTOR_H_

#include <iostream>
using std::ostream ;
using std::istream ;

#include "Functions.H" 
#include "Const_num.H"

class Matrix ;

class Vector
{
  friend ostream &operator<<( ostream &, const Vector & ) ;
  friend istream &operator>>( istream &, Vector & ) ;
public:
  Vector() ;                  // default constructor
  Vector(int) ;            // set all elements to zero     
  Vector(const Vector &) ;    // copy constructor
  virtual ~Vector() ;                 // destructor

  // test if vector is null
  bool IsNull() const ;   // return true if it is null false otherwise
 
  int Size() const ;          // return size

  // this function set the size of a previously declared vector.
  // set all the elements to zero. Vector must be uninitialized.
  void SetSize(int) ;

  // This function clear a vector, that is, transforming it to a non
  // initialized vector. It has the same effect that the destructor. 
  void Clear() ;

  // this function return a const pointer to the vector data.
  // It is mainly intended to be used inside SquareMatrix class.
  // this is the const overloaded function.
  const double *Data() const;  

  // this function return a pointer to the vector data.
  // It is mainly intended to be used inside SquareMatrix class.
  // Note that this function could be used as a modifiable lvalue and
  // that, if is use in a wrong way, can cause problems 
  double *Data() ; 

  const Vector &operator=( const Vector & ) ;  // assigment operator
 
  // subscript operator for non-const objects returns modifiable lvalue
  double &operator()( int ) ;

  // subscript operator for const objects returns rvalue
  double operator()( int ) const ;

  // set all the elements of vector v to the value x
  // v_i = x
  void SetAll( double ) ;

  // sets all the elements to the value zero
  // v_i = 0
  void SetZero() ;

  // makes a basis vector by setting all the elements of the vector
  // equal to zero except for the i-th element, which is set to one
  void SetBasis( int ) ;

  // multiplies the elements of vector a by the constant factor x.
  // a'_i = x * a_i
  void Scale( double ) ;

  // adds the constant value x to the elements of the vector a.
  // a'_i = a_i + x
  void AddConstant( double ) ;

  // += overloded operator for non-const objects.
  // a'_i = a_i + b_i
  Vector &operator+=( const Vector & ) ;

  // -= overloded operator for non-const objects.
  // a'_i = a_i - b_i
  Vector &operator-=( const Vector & ) ;

  // *= overloded operator for non-const objects.
  // a'_i = a_i * b_i
  Vector &operator*=( const Vector & ) ;

  // this function returns the maximum value in the vector.
  double MaxValue() const ;

  // this function returns the minimum value in the vector.
  double MinValue() const ;

  // this function returns the minimim and the maximum values in
  // the vector v, storing in min_out and max_out.
  void MinMaxValue( double &min_out, double &max_out) const ;

  // this function returns the index of the maximum value in the vector v.
  // when there are several maximum elements then the lowest index is
  // returned
  int MaxIndex() const ;

  // this function returns the index of the minimum value in the vector v.
  // when there are several minimum elements then the lowest index is
  // returned
  int MinIndex() const ;

  // this function returns the index of the maximum and minimum values
  // in the vector v, storing them in imin and imax. when there are several 
  // maximum or minimum elements then the lowest indeces are returned
  void MinMaxIndex( int &imin, int &imax ) const ;

  // this function compute the Euclidean norm ||x||_2 = \sqrt{\sum x_i^2}
  double EuclideanNorm() const ;

  // this function compute the absolute sum. \sum |x_i| of the elements
  // of the vector x
  double AbsoluteSum() const ;

  // this function compute the scalar product x^T y for the vector x and y
  // i.e. x.DotProduct(y) = x^T y
  double DotProduct( const Vector & ) const ;

  // this function compute the matrix-vector product and sum y' = A x + y
  // i.e. Vector x(2) ; x.SetAll(1.0) ; Matrix A(2,2) ;
  //      A.setAll(2.0) ; Vector y(2) ; y.SetAll(3.0) ; y.dgemv(A,x) ; 
  void dgemv(const Matrix &A, const Vector &x) ;

private:
  int _size ;              // pointer based array size
  double *_data ;          // pointer to first element of pointer-based array 
} ; // end class Vector

#endif
