/****************************************************************************
 * Error handler and ran1 function                                          *
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 *  unkown     Dr. Simkins    Original code                                 *
 * 07/22/09    Mario Juha     copy from original code                       *
 * 07/08/10    Mario Juha     Added ran2 function                           *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#include <iostream>
using std::cerr ;
using std::endl ;

#include <cstdlib>  // contains exit function prototype
using std::exit ;

#include "Const_num.H"

//-----------------

void error(const char *routine, const char *message)
{
  cerr << routine << ": ERROR: " << message << endl ;
  cerr << "Error handler invoked." << endl ;
  cerr << "Aborted" << endl ;
  exit(1) ;
}

//------------------

void warning(const char *routine, const char *message)
{
  cerr << routine << ": WARNING: " << message << endl ;
  cerr << "Check results carefully!" << endl ;
}

//------------------

// Long period (> 2 x 10^18) random number generator of L'Ecuyer with
// Bays-Durham shuffle and added safeguards. Returns a uniform random
// deviate between 0.0 and 1.0 (exclusive of the endpoint
// values). Call with idum a negative integer to initialize;
// thereafter, do not alter idum between successive deviates in a
// sequence. RNMX should approximate the largest floating value that
// is less than 1.
//
// W. Press, S. Teukolsky, W. Vetterling and B. Flannery. Numerical
// Recipies in C. The Art of Scientific Computing. Second
// Edition. Cambrige University Press. 2002. Page 282.
float ran2(long &idum)
{
  int j ;
  long k ;
  static long idum2 = 123456789 ;
  static long iy = 0 ;
  static long iv[NTAB] ;
  float temp ;

  if (idum <= 0 )
    {
      // initialize
      if (-idum < 1 ) idum = 1 ; // Be sure to prevent idum = 0.
      else idum = -idum ;
      idum2 = idum ;
      for (j=NTAB+7;j>=0;j--)
	{
	  // Load the shuffle table (after 8 warm-ups)
	  k=idum/IQ1 ;
	  idum=IA1*(idum-k*IQ1)-k*IR1 ;
	  if (idum < 0 ) idum += IM1 ;
	  if (j < NTAB) iv[j] = idum ;
	}
      iy=iv[0] ;
    }
  k=idum/IQ1 ; // Start here when not initializing
  idum=IA1*(idum-k*IQ1)-k*IR1 ; // Compute idum=(IA1*idum) % IM1
				// without over flows by Schrage's
				// method.
  if (idum < 0) idum += IM1 ;
  k=idum2/IQ2 ;
  idum2 = IA2*(idum2-k*IQ2)-k*IR2 ; // Compute idum2=(IA2*idum) % IM2
				    // likewise.
  if (idum2 < 0) idum2 += IM2 ;
  j=iy/NDIV ; // Will be in the range 0..NTAB-1
  iy=iv[j]-idum2 ; // Here idum is shuffled, idum and idum2 are
		   // combined to generate output.
  iv[j] = idum ;
  if (iy < 1) iy += IMM1 ;
  if ((temp=AM*iy) > RNMX) return RNMX ; // Because users don't expect
					 // endpoint values.
  else return temp ;
}
