/****************************************************************************
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#include <iostream>
using std::cerr ;
using std::cout ;
using std::cin ;
using std::endl ;
#include <cstring>
#include <algorithm>
using std::min ;
using std::max ;

#include "Vector.H"  // Vector class definition
#include "Matrix.H"  // Matrix class definition

//-----------------------
// default constructor for class Matrix
Matrix::Matrix()
{
  _data = 0 ;      // point to null
  _rows =  0 ;     // set number of rows equal to zero
  _cols = 0 ;   // set number of columns equal to zero
  _status = UNINITIALIZED ;
} 

//-----------------------
// Parametrized constructor for class Matrix
Matrix::Matrix( int rows , int cols )
{
  if ( ( rows > 0 ) && ( cols > 0 ) )
    {
      // allocate space for matrix and initialize all element to zero
      const int N = rows*cols ;
      _data = new double [N] ;
      memset(_data,0,N*sizeof(double)) ;
      _rows = rows ;
      _cols = cols ;
      _status = NORMAL ;  
    }
  else
    {
      error("Matrix.C: In function 'Matrix::Matrix(...)'",
	    "Matrix length must be a positive integer")  ;
    }
}

//-----------------------
// copy constructor for class Matrix
// must receive a reference to prevent infinite recursion
Matrix::Matrix( const Matrix &MatrixToCopy )
{
  const double *src = MatrixToCopy._data ;
  _rows = MatrixToCopy._rows ;  // get number or rows
  _cols = MatrixToCopy._cols ;  // get number or cols
  _status = MatrixToCopy._status ;  // get status
  if ( ( _rows == 0 ) || ( _cols == 0 ) )
    {
      error("Matrix.C: In function 'Matrix::Matrix(Matrix)'",
  	    "Matrix length must be a positive integer") ;
    }
  const int N = _rows * _cols ;
  _data = new double [N] ;
  cblas_dcopy(N,src,1,_data,1) ;
}

//-----------------------
// destructor for class Matrix
Matrix::~Matrix()
{
  if(_data != 0)
    {
      delete [] _data ;
      _data = 0 ;
    }
  _rows = 0 ;
  _cols = 0 ;
  _status = UNINITIALIZED ;
}

//-----------------------

bool Matrix::IsNull() const
{
  return ( _status == UNINITIALIZED  ) ;
}

//-----------------------

int Matrix::Rows() const
{
  return _rows ;  // number of rows in matrix
}

//-----------------------

int Matrix::Cols() const
{
  return _cols ;  // number of columns in matrix
}

//-----------------------
// this function returns a constant pointer to the matrix data.
const double *Matrix::Data() const
{
  return _data ;
}  

//-----------------------
// this function return the status of the matrix; encoding using the
// variable Status defined above
Status Matrix::GetStatus() const
{
  return _status ;
}

//-----------------------

// This function clear a matrix, that is transforming it to a non
// initialized matrix. It has the same effect that the destructor. 
void Matrix::Clear()
{
  if(_data != 0)
    {
      delete [] _data ;
      _data = 0 ;
    }
  _rows = 0 ;
  _cols = 0 ;
  _status = UNINITIALIZED ;
}

//-----------------------
// this function set the number of rows and columns in matrix m.
// The matrix must be uninitialized.
// i.e. Matrix m ; m.SetSize(2,2) ;
void Matrix::SetSize(int row, int col)
{
  if ( _status != UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::SetSize()'",
	    "Matrix was initialized!") ;
    }
  else if ( ( row <= 0 ) || ( col <= 0 ) )
    {
      error("Matrix.C: In function 'Matrix::SetSize()'",
	    "Matrix length must be a positive integer") ;
    }
  else
    {
      _rows = row ;
      _cols = col ;
      _data = new double [row*col] ; // allocate space for matrix
      _status = NORMAL ;
    }
}

//-----------------------
// this function set the status of the matrix; encoding using the
// variable Status defined above
void Matrix::SetStatus(Status status)
{
  _status = status ;
} 

//-----------------------
// this function return a reference to the matrix data.
// It is mainly intended to be used inside SquareMatrix class.
// Note that this function could be using like a modifiable lvalue and
// that if is use in a wrong way can cause problems 
double *Matrix::Data()
{
  return _data ;
} 

//-----------------------
// overloaded assigment operator ;
// const return avoids: (a1 = a2) = a3
const Matrix &Matrix::operator=( const Matrix &right )
{
  if ( &right != this ) // avoid self-assigment
    {
      if (  ( _rows != right._rows ) ||
	    ( _cols != right._cols ) )
	{
	  error("Matrix.C: In function '&Matrix::operator='",
		"Matrix has different sizes") ;
	}
      else if ( ( _rows == 0 ) || ( _cols == 0 ) )
	{
	  error("Matrix.C: In function '&Matrix::operator='",
		"Matrix length must be a positive integer") ;
	}
      else
	{
	  const double *src = right._data ;
	  _rows = right._rows ;
	  _cols = right._cols ;
	  _status = right._status ;
	  const int N = _rows * _cols ;
	  cblas_dcopy(N,src,1,_data,1) ; //copy one matrix into another
	}
    }
  return *this ;  // enables x = y = z, for example
}

//-----------------------
// overloded subscript operator for non-const Matrix;
// reference return creates a modifiable lvalue
double &Matrix::operator()( int row , int col )
{
  if ( ( row < 0 ) || ( row >= _rows ) 
       || ( col < 0 ) || ( col >= _cols ) )
    {
      error("Matrix.C: In function '&Matrix::operator()'",
	    "Index out of range") ;
    }
  return _data[_rows*col+row] ;  // reference return
}

//-----------------------
// overloded subscript operator for const Matrix;
// const reference return creates an rvalue
double Matrix::operator()( int row , int col ) const
{
  if ( ( row < 0 ) || ( row >= _rows ) 
       || (  col < 0 ) || ( col >= _cols ) )
    {
      error("Matrix.C: In function 'Matrix::operator()'",
	    "Index out of range") ;
    }
  return _data[_rows*col+row] ;  // copy return
}

//-----------------------
// sets all the elements to the value of x
void Matrix::SetAll( double x )
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::SetAll(...)'",
	    "Matrix uninitialized!") ;
    }
  const int N = _rows * _cols ;
  for (int i=0; i<N; i++) _data[i] = x ;
  _status = NORMAL ;
}

//-----------------------
// sets all the elements to the value zero
void Matrix::SetZero()
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::SetZero()'",
	    "Matrix uninitialized!") ;
    }
  const int N = _rows*_cols ;
  memset(_data,0,N*sizeof(double)) ; // set all element to zero
  _status = NORMAL ;
}

//-----------------------
// multiplies the elements of matrix a by the constant factor x.
// a'(i,j) = x * a(i,j)
void Matrix::Scale( double x )
{
  if (_status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::Scale(...)'",
	    "Matrix uninitialized!") ;
    }
  const int N = _rows*_cols ;
  cblas_dscal(N,x,_data,1) ;
  _status = NORMAL ;
}

//-----------------------
// += overloded operator for non-const objects.
// a = a + b
Matrix &Matrix::operator+=( const Matrix &right )
{
  if (  ( _rows != right._rows ) ||
	( _cols != right._cols ) )
    {
      error("Matrix.C: In function '&Matrix::operator+='",
	    "Matrix has different sizes") ;
    }
  else if ( ( _rows == 0 ) || ( _cols == 0 ) )
    {
      error("Matrix.C: In function '&Matrix::operator+='",
	    "Matrix length must be a positive integer") ;
    }
  else
    {
      const int N = _rows*_cols ;
      cblas_daxpy(N,one,right._data,1,_data,1) ;
    }
  return *this ;  // return a reference to this element
}

//-----------------------
// -= overloded operator for non-const objects.
// a = a - b
Matrix &Matrix::operator-=( const Matrix &right ) 
{
  if (  ( _rows != right._rows ) ||
	( _cols != right._cols ) )
    {
      error("Matrix.C: In function '&Matrix::operator-='",
	    "Matrix has different sizes") ;
    }
  else if ( ( _rows == 0 ) || ( _cols == 0 ) )
    {
      error("Matrix.C: In function '&Matrix::operator-='",
	    "Matrix length must be a positive integer") ;
    }
  else
    {
      const int N = _rows*_cols ;
      cblas_daxpy(N,-one,right._data,1,_data,1) ;
    }
  return *this ;  // return a reference to this element
}

//-----------------------
// *= overloded operator for non-const objects.
// a'(i,j) = a(i,j) * b(i,j)
Matrix &Matrix::operator*=( const Matrix &right )
{
  if (  ( _rows != right._rows ) ||
	( _cols != right._cols ) )
    {
      error("Matrix.C: In function '&Matrix::operator*='",
	    "Matrix has different sizes") ;
    }
  else if ( ( _rows == 0 ) || ( _cols == 0 ) )
    {
      error("Matrix.C: In function '&Matrix::operator*='",
	    "Matrix length must be a positive integer") ;
    }
  else
    {
      for (int i=0; i<_rows; i++)
	{
	  for (int j=0; j<_cols; j++)
	    {
	      _data[_rows*j+i] = _data[_rows*j+i] * _data[_rows*j+i] ;
	    }
	}
    }
  return *this ;  // return a reference to this element
}

//-----------------------
// this function returns the maximum value in the matrix.
double Matrix::MaxValue() const
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::MaxValue()'",
	    "Matrix uninitialized or empty") ;
    }
  const int N = _rows * _cols ;
  int index = cblas_idamax(N,_data,1) ;
  return _data[index] ;
}

//-----------------------
// this function returns the minimum value in the matrix.
double Matrix::MinValue() const
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::MinValue()'",
	    "Matrix uninitialized or empty") ;
    }
  
  // Took and modified from GNU gsl
  /* finds the smallest element of a matrix */
  double min_val = bigreal ;
  for (int i = 0; i < _rows; i++)
    {
      for (int j = 0; j < _cols; j++)
	{
	  double x = _data[_rows*j + i] ;
	  min_val = min(x,min_val) ;
	}
    }

  return min_val ;
}

//-----------------------
// this function returns the minimim and the maximum values in
// the matrix m, storing in min_out and max_out.
void Matrix::MinMaxValue( double &min_out, double &max_out) const
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::MinMaxValue()'",
	    "Matrix uninitialized or empty") ;
    }
  // Took and modified from GNU gsl
  // min max
  double min_val = bigreal ;
  double max_val = -bigreal ;
  for (int i = 0; i < _rows; i++)
    {
      for (int j = 0; j < _cols; j++)
	{
	  double x = _data[_rows*j + i] ;
	  min_val = min(min_val,x) ;
	  max_val = max(max_val,x) ;
	}
    }
  // return values
  min_out = min_val ;
  max_out = max_val ;
}

//-----------------------
// this function returns the index of the maximum value in the matrix m.
// when there are several maximum elements then the lowest index is
// returned
void Matrix::MaxIndex( int &imax , int &jmax ) const
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::MaxIndex()'",
	    "Matrix uninitialized or empty") ;
    }
  // Took and modified from GNU gsl
  double max_val = -bigreal ;
  for (int i = 0; i < _rows; i++)
    {
      for (int j = 0; j < _cols; j++)
	{
	  double x = _data[_rows*j+i] ;
	  if (x > max_val)
            {
              max_val = x;
              imax = i;
              jmax = j;
            }
	}
    }
}

//-----------------------
// this function returns the index of the minimum value in the matrix m.
// when there are several minimum elements then the lowest index is
// returned
void Matrix::MinIndex( int &imin , int &jmin ) const
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::MinIndex()'",
	    "Matrix uninitialized or empty") ;
    }
  // Took and modified from GNU gsl
  double min_val = bigreal ;
  for (int i = 0; i < _rows; i++)
    {
      for (int j = 0; j < _cols; j++)
	{
	  double x = _data[_rows*j+i] ;
	  if (x < min_val)
            {
              min_val = x;
              imin = i;
              jmin = j;
            }
	}
    }
}

//-----------------------
// this function returns the index of the maximum and minimum values
// in the matrix m, storing them in imin and imax. when there are several 
// maximum or minimum elements then the lowest indeces are returned
void Matrix::MinMaxIndex( int &imin, int &jmin ,
			  int &imax, int &jmax ) const
{
  if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::MinMaxIndex()'",
	    "Matrix uninitialized or empty") ;
    }
  // Took and modified from GNU gsl
  double min_val = bigreal ;
  double max_val = -bigreal ;
  for (int i = 0; i < _rows; i++)
    {
      for (int j = 0; j < _cols; j++)
	{
	  double x = _data[_rows*j+ i];
	  if (x < min_val)
	    {
	      min_val = x;
	      imin = i;
	      jmin = j;
	    }
	  if (x > max_val)
	    {
	      max_val = x;
	      imax = i;
	      jmax = j;
	    }
	}
    }
}

//-----------------------
// this function copies the elements of the i-th row of the matrix m
// into the vector v.
void Matrix::GetRow(int row, Vector &dest ) const
{
  if ( ( row < 0 ) || ( row >= _rows ) )
    {
      error("Matrix.C: In function 'Matrix::GetRow(...)'",
	    "Index out of range") ;
    }
  else if ( dest.Size() != _cols )
    { 
      error("Matrix.C: In function 'Matrix::GetRow(...)'",
	    "Vector does not have compatible size with matrix cols") ;
    }
  else if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::GetRow(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else
    {
      for (int j = 0; j < _cols; j++)
	dest(j) = _data[_rows*j+row] ;
    }
}

//-----------------------
// this function copies the elements of the j-th col of the matrix m
// into the vector v.
void Matrix::GetCol(int col, Vector &dest ) const
{
  if ( ( col < 0 ) || ( col >= _cols ) )
    {
      error("Matrix.C: In function 'Matrix::GetCol(...)'",
	    "Index out of range") ;
    }
  else if ( dest.Size() != _rows )
    { 
      error("Matrix.C: In function 'Matrix::GetCol(...)'",
	    "Vector does not have compatible size with matrix rows") ;
    }
  else if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::GetCol(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else
    {
      for (int i = 0; i < _rows; i++)
	dest(i) = _data[_rows*col+i] ;
    }
}

//-----------------------
// this function copies the elements of the vector v into the i-th row 
// of the matrix m.
void Matrix::SetRow(const Vector &src, int row)
{
  if ( ( row < 0 ) || ( row >= _rows ) )
    {
      error("Matrix.C: In function 'Matrix::SetRow(...)'",
	    "Index out of range") ;
    }
  else if ( src.Size() != _cols )
    { 
      error("Matrix.C: In function 'Matrix::SetRow(...)'",
	    "Vector does not have compatible size with matrix cols") ;
    }
  else if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::SetRow(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else
    {
      for( int j=0; j<_cols; j++)
	_data[_rows*j+row] = src(j) ;
      
      _status = NORMAL ;
    }
}

//-----------------------
// this function copies the elements of the vector v into the j-th col 
// of the matrix m.
void Matrix::SetCol(const Vector &src, int col)
{
  if ( ( col < 0 ) || ( col >= _cols ) )
    {
      error("Matrix.C: In function 'Matrix::SetCol(...)'",
	    "Index out of range") ;
    }
  else if ( src.Size() != _rows )
    { 
      error("Matrix.C: In function 'Matrix::SetCol(...)'",
	    "Vector does not have compatible size with matrix rows") ;
    }
  else if ( _status == UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::SetCol(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else
    {
      for( int i=0; i<_rows; i++)
	_data[_rows*col+i] = src(i) ;

      _status = NORMAL ;
    }
}

//-----------------------
// this function makes the matrix dest the transpose of the matrix src by
// copying the elements of src into dest. This function works for all
// matrices provided that the matrix dest is uninitialized.
// i.e. Matrix dest ; src.Transpose(dest) ;
void Matrix::Transpose(Matrix &dest) const
{
  if ( dest._status != UNINITIALIZED )
    {
      error("Matrix.C: In function 'Matrix::Transpose(...)'",
	    "Matrix was initialized!") ;
    }
  dest.SetSize(_cols,_rows) ;
  for (int i=0; i<_rows; i++)
    {
      for (int j=0; j<_cols; j++)
	{
	  dest(j,i) = _data[_rows*j+i] ; 
	}
    }
  dest._status = NORMAL ;
}

//-----------------------
// this function compute the matrix-matrix product and sum C' = A B + C
// i.e. Matrix A(2,3) ; A.SetAll(1.0) ; Matrix B(3,2) ;
//      B.setAll(2.0) ; Matrix C(2,2) ; C.SetAll(3.0) ; C.dgemm(A,B) ; 
void Matrix::dgemm(const Matrix &A, const Matrix &B) const
{
  if ( ( A._status == UNINITIALIZED ) || 
       ( B._status == UNINITIALIZED ) ||
       ( _status == UNINITIALIZED ) )
    {
      error("Matrix.C: In function 'Matrix::dgemm(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else if ( ( A._rows != _rows ) || ( _cols != B._cols ) )
    {
      error("Matrix.C: In function 'Matrix::dgemm(...)'",
	    "Matrix-matrix does not match the sizes for summation") ;
    }
  else if ( A._cols != B._rows )
    {
      error("Matrix.C: In function 'Matrix::dgemm(...)'",
	    "Matrix-matrix does not match the sizes for multiplication") ;
    }
  else
    {
      cblas_dgemm(CblasColMajor,CblasNoTrans,CblasNoTrans,A._rows,B._cols,
		  A._cols,one,A._data,A._rows,B._data,B._rows,one,_data,
		  _rows) ;
    }
}

//-----------------------
// Computes the minimum norm solution to a real linear least
// squares problem:
// Minimize 2-norm(| b - A*x |).
// using the singular value decomposition (SVD) of A. A is an M-by-N
// matrix which may be rank-deficient.
// The condition number of A in the 2-norm = S(1)/S(min(m,n)).
// where S are the singular values of A.
void Matrix::SVD_Solve(Vector &x,double &cond_number)
{
  if ( x.IsNull() )
    {
      error("Matrix.C: In function 'Matrix::SVD_Solve(...)'",
  	    "Left vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("Matrix.C: In function 'Matrix::SVD_Solve(...)'",
  	    "Matrix is null") ;
    }
  else if ( x.Size() != Cols() ) // call base-clase Cols() function
    {
      error("Matrix.C: In function 'Matrix::SVD_Solve(...)'",
  	    "Matrix-vector size does not match") ;
    }
  else
    {
      const int M = _rows ;
      const int N = _cols ; 
      const int lda = _rows ;
      const int ldb = 1 ;
      const int NRHS = 1 ;
      int lwork = 3*min(M,N) + max(2*min(M,N),max(M,N)) ;
      double *work = new double [ max(1,lwork) ] ;
      double *S = new double [ min(M,N) ] ;
      int rank, info ;
      const double rcond = ZeroTol ;

      // compute SVD decomposition
      dgelss_(&M,&N,&NRHS,_data,&lda,x.Data(),&ldb,S,&rcond,&rank,work,&lwork,&info) ;

      // check output
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument had an illegal value." << endl ;
	  error("Matrix.C: In function 'Matrix::SVD_Solve(...)'",
  		"Illegal argument in dgesdd_, please check!") ; 
	}
      else if ( info > 0 )
	{
	  error("Matrix.C: In function 'Matrix::SVD_Solve(...)'",
  		"the algorithm for computing the SVD failed to converge") ;
	}
      else
	{
	  const int index = min(M,N) - 1 ;
	  const double min_val = S[index] ;
	  const double max_val = S[0] ;
	 if ( min_val <= ZeroTol )
	   {
	     error("Matrix.C: In function 'Matrix::SVD_Solve(...)'",
		   "Matrix is singular") ; 
	   }
	 else
	   {
	     cond_number = max_val / min_val ;
	   } 
	}
      // cleaning up memory
      delete [] S ;
      S = 0 ;
      delete [] work ;
      work = 0 ;
      // set status
      SetStatus(SVD_DECOMP) ;
    }
}

//-----------------------
// overloded output operator for class Matrix
ostream &operator<<( ostream &output, const Matrix &a )
{
  if ( (a._status == UNINITIALIZED) )//|| (a._status == EMPTY) )
    {
      error("Matrix.C: In function '&operator<<(...)'",
	    "Matrix uninitialized or empty") ;
    }

  // output private ptr-based array
  for ( int i = 0; i < a._rows; i++ )
    {
      for (int j = 0; j < a._cols; j++ )
	{
	  output << a(i,j) << ' ' ;
	  if ( ( j + 1 ) % a._cols == 0 )
	    output << endl ;
	}
    }

  return output ; // enables cout << x << y
}

//-----------------------
// overloded input operator for class Matrix
istream &operator>>( istream &input, Matrix &a )
{
  if ( (a._status == UNINITIALIZED) )//|| (a._status == EMPTY) )
    {
      error("Matrix.C: In function '&operator>>(...)'",
	    "Matrix uninitialized or empty") ;
    }

  // read array
  for ( int i = 0; i < a._rows; i++ )
    {
      for (int j = 0; j < a._cols; j++ )
	{
	  input >> a(i,j) ;
	}
    }

  return input ; // enables cin >> A >> B
}

//-----------------------

