/****************************************************************************
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 * 09/30/10    Mario Juha   Added the overloading >> and << operators.      *
 *                          A bug fixed on LinSolveExpert.                  *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#include <iostream>
using std::cerr ;
using std::cout ;
using std::cin ;
using std::endl ;

#include "Const_num.H" 
#include "Vector.H"  // Vector class definition
#include "Matrix.H"  // Matrix class definition
#include "SparseMatrix.H"  // SparseMatrix class definition

//-----------------------
// default constructor for class SparseMatrix
SparseMatrix::SparseMatrix()
{
  _size = 0 ;      // set number of rows and cols equal to zero
  _nnz = 0 ;       // set number of nonzeros elements in the matrix
  _status = UNINITIALIZED ;
  _nzval = 0 ;  // point to null
  _rowind = 0 ; // point to null
  _colptr = 0 ; // point to null
} 

//-----------------------
// Parametrized constructor for class SparseMatrix
SparseMatrix::SparseMatrix( int size, int nnz)
{
  if ( ( size > 0 ) && ( nnz > 0 ) )
    {
      _size = size ;
      _nnz = nnz ;
      _status = NORMAL ; 
      
      // initialize arrays
      _nzval = new double [_nnz] ;
      _rowind = new int [_nnz] ;
      _colptr = new int [_size+1] ;
    }
  else
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SparseMatrix(...)'",
	    "SparseMatrix length must be a positive integer")  ;
    }
}

//-----------------------
// Parametrized constructor 
SparseMatrix::SparseMatrix( int size, int nnz, double *a, int *asub, int *xa)
{
  if ( ( size > 0 ) && ( nnz > 0 ) 
       && (a != 0) && (asub != 0) && (xa != 0) )
    {
      _size = size ;
      _nnz = nnz ;
      _status = NORMAL ; 
      // pointer to arrays
      _nzval = a ;
      _rowind = asub ;
      _colptr = xa ;
    }
  else
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SparseMatrix(...)'",
	    "SparseMatrix length must be a positive integer")  ;
    }
}
//-----------------------

//-----------------------
// copy constructor for class SparseMatrix
// must receive a reference to prevent infinite recursion
SparseMatrix::SparseMatrix( const SparseMatrix &SparseMatrixToCopy )
{
  _size = SparseMatrixToCopy._size ;  // get number of rows
  _nnz = SparseMatrixToCopy._nnz ;  // get number of nonzero elements
  _status = SparseMatrixToCopy._status ;  // get status
  if ( ( _size == 0 ) || ( _nnz == 0 ) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SparseMatrix(...)'",
  	    "SparseMatrix length must be a positive integer") ;
    }
  // allocate memory
  _nzval = new double [_nnz] ;
  _rowind = new int [_nnz] ;
  _colptr = new int [_size+1] ;

  if ( ( SparseMatrixToCopy._nzval != 0 ) &&
       ( SparseMatrixToCopy._rowind != 0 ) &&
       ( SparseMatrixToCopy._colptr != 0 ) )
    {
      // copy _nzval array
      cblas_dcopy( _nnz, SparseMatrixToCopy._nzval, 1, _nzval, 1 );
      // copy _rowind array
      memcpy( _rowind, SparseMatrixToCopy._rowind, _nnz*sizeof(int) ) ;
      // copy _colptr array
      memcpy( _colptr, SparseMatrixToCopy._colptr, (_size + 1)*sizeof(int) ) ;
    }
    else
      {
	error("SparseMatrix.C: In function 'SparseMatrix::SparseMatrix(...)'",
	      "_nzval, _rowind or _colptr point to null!") ;
      }
}

//-----------------------
// destructor for class SparseMatrix
SparseMatrix::~SparseMatrix()
{
  if ( _nzval != 0 )
    {
      delete [] _nzval ;
      _nzval = 0 ;
    }
  if ( _rowind != 0 )
    {
      delete [] _rowind ;
      _rowind = 0 ;
    }
  if ( _colptr != 0 )
    {
      delete [] _colptr ;
      _colptr = 0 ;
    }
  _size = 0 ;      // set number of rows equal to zero
  _nnz = 0 ;       // set number of nonzeros elements in the matrix
  _status = UNINITIALIZED ;
}

//-----------------------

bool SparseMatrix::IsNull() const
{
  return ( _status == UNINITIALIZED  ) ;
}

//-----------------------

int SparseMatrix::Size() const
{
  return _size ;  // number of rows in matrix
}

//-----------------------


//-----------------------
int SparseMatrix::NumberNonZeros() const 
{
  return _nnz ;  // return number of non zero values 
}
//-----------------------

//-----------------------
// sets all the elements to the value zero
// m(i,j) = 0
void SparseMatrix::SetZero()
{
  if ( _status != UNINITIALIZED )
    {
      memset(_nzval,0,_nnz*sizeof(double)) ;
    }
  else
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetZero()'",
 	    "SparseMatrix uninitialized!") ;
    }
}

//-----------------------
// this function set the nonzero coefficients of matrix A into array
// of _nzval.
void SparseMatrix::SetNZValues(int index, double val)
{
  if ( _nnz == 0 )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetNZValues(...)'",
	    "_nnz is zero!") ;
    }
  else if ( (index < 0 ) || (index >= _nnz) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetNZValues(...)'",
	    "Index out of range!") ;
    }
  else
    {
      _nzval[index] = val ;
    }
}

//-----------------------

// this function set the row indices
void SparseMatrix::SetRowIndex(int index, int val )
{
  if ( _nnz == 0 )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetRowIndex(...)'",
	    "_nnz is zero!") ;
    }
  else if ( (index < 0 ) || (index >= _nnz) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetRowIndex(...)'",
	    "Index out of range!") ;
    }
  else
    {
      _rowind[index] = val ;
    }
}

//-----------------------
// this function set the beginning of each column in the coefficient
// and row index array.
void SparseMatrix::SetColIndex(int index, int val)
{
  if ( _size == 0 )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetColIndex(...)'",
	    "_size is zero!") ;
    }
  else if ( (index < 0 ) || (index > _size) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetColIndex(...)'",
	    "Index out of range!") ;
    }
  else
    {
      _colptr[index] = val ;
    }
}

//-----------------------
// this function return a const pointer to the row index data.
const int *SparseMatrix::GetRowIndexArray() const
{
  return _rowind ;
}

//-----------------------
// this function return a pointer to the row index data.
// Note that this function could be used as a modifiable lvalue and
// that, if is use in a wrong way, can cause problems 
int *SparseMatrix::GetRowIndexArray()
{
  return _rowind ;
} 

//-----------------------
// this function return a const pointer to the col index data.
const int *SparseMatrix::GetColIndexArray() const
{
  return _colptr ;
} 

//-----------------------
// this function return a pointer to the col index data.
// Note that this function could be used as a modifiable lvalue and
// that, if is use in a wrong way, can cause problems 
int *SparseMatrix::GetColIndexArray()
{
  return _colptr ;
}

//-----------------------
// subscript operator for non-const objects returns modifiable lvalue
double &SparseMatrix::operator()( int row, int col )
{
  if ( ( row < 0 ) || ( row >= _size ) 
       || ( col < 0 ) || ( col >= _size ) )
    {
      error("SparseMatrix.C: In function '&SparseMatrix::operator()'",
	    "Index out of range") ;
    }
  int istart = _colptr[col] ;
  int iend = _colptr[col+1] ;
  int index = 0 ;

  bool flag = true ;
  for ( int k=istart; k<iend; k++)
    {
      if ( _rowind[k] == row )
	{
	  index = k ;
	  flag = false ;
	  break ;
	}
    }
  if ( flag )
    {
      error("SparseMatrix.C: In function '&SparseMatrix::operator()'",
	    "Trying to access or modify a forbiden location!") ;
    }
  return _nzval[index] ;  // reference return
}

//-----------------------
// subscript operator for const objects returns rvalue
double SparseMatrix::operator()( int row , int col ) const
{
  if ( ( row < 0 ) || ( row >= _size ) 
       || ( col < 0 ) || ( col >= _size ) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::operator()'",
	    "Index out of range") ;
    }
  int istart = _colptr[col] ;
  int iend = _colptr[col+1] ;
  int index = 0 ;

  bool flag = true ;
  for ( int k=istart; k<iend; k++)
    {
      if ( _rowind[k] == row )
	{
	  index = k ;
	  flag = false ;
	  break ;
	}
    }
  if ( flag )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::operator()'",
	    "Trying to access a forbiden location!") ;
    }
  return _nzval[index] ;  // reference return
}

//-----------------------
// This function computes the solution to a real system of linear
// equations A * x = b, using SuperLU library.  where A is an N-by-N
// sparse (Harwell-Boing format) matrix and x and b are N-by-1
// vectors.
//
//  The LU decomposition with partial pivoting and row interchanges is
//  used to factor A as
//     A = P * L * U,
//  where P is a permutation matrix, L is unit lower triangular, and U is
//  upper triangular.  The factored form of A is then used to solve the
//  system of equations A * x = b.
void SparseMatrix::LinSolve(Vector &x) const
{
  if ( x.IsNull() )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::LinSolve(...)'",
	    "Left vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SparseMatrix.C: In function 'SparseMatrix::LinSolve(...)'",
	    "Matrix is null") ;
    }
  else if ( x.Size() != _size ) // call base-clase Size() function
    {
      error("SparseMatrix.C: In function 'SparseMatrix::LinSolve(...)'",
	    "Matrix-vector size does not match") ;
    }
  else
    {
      SuperMatrix A, B, L, U ;
      // create matrix A in the format expected by SuperLU
      dCreate_CompCol_Matrix(&A,_size,_size,_nnz,_nzval,_rowind,_colptr,
			     SLU_NC,SLU_D,SLU_GE) ;

      // create right-hand side matrix B
      const int nrhs = 1 ; // number of right-hand side
      dCreate_Dense_Matrix(&B,_size,nrhs,x.Data(),_size,SLU_DN,SLU_D,SLU_GE) ;

      // row permutations from partial pivoting
      int *perm_r = new int [_size] ;
      // column permutation vector
      int *perm_c = new int [_size] ;

      // set default input options
      superlu_options_t options ;
      set_default_options(&options) ;
      options.PrintStat = NO ;

      // initialize the statistics variables
      SuperLUStat_t stat ;
      StatInit(&stat) ;

      // solve equations
      int info ;
      dgssv(&options,&A,perm_c,perm_r,&L,&U,&B,&stat,&info) ;

      // checking output
      if ( info > 0 )
	{
	  if ( info <= _size )
	    {
	      cerr << "U( " << info << " , " << info << " ) is exactly zero" << endl ;
	      error("SparseMatrix.C: In function 'SparseMatrix::LinSolve(...)'",
		    "Solution could not be computed, factor U is exactly singular") ;
	    }
	  else
	    {
	      error("SparseMatrix.C: In function 'SparseMatrix::LinSolve(...)'",
		    "Memory allocation failure occured") ;
		    
	    }
	}

      // de-allocate storage
      delete [] perm_r ;
      perm_r = 0 ;
      delete [] perm_c ;
      perm_c = 0 ;
      // destroy SuperMatrix
      Destroy_SuperMatrix_Store(&A) ;
      Destroy_SuperMatrix_Store(&B) ;
      Destroy_SuperNode_Matrix(&L) ;
      Destroy_CompCol_Matrix(&U) ;
      StatFree(&stat) ;
    }
} 

//-----------------------
// This function computes the solution to a real system of linear
// equations A * x = b, using SuperLU library.  where A is an N-by-N
// sparse (Harwell-Boing format) matrix and x and b are N-by-1
// vectors.
//
//  The LU decomposition with partial pivoting and row interchanges is
//  used to factor A as
//     A = P * L * U,
//  where P is a permutation matrix, L is unit lower triangular, and U
//  is upper triangular.  The factored form of A is then used to solve
//  the system of equations A * x = b. An expert driver algorithm is
//  used and the inverse of the condition number and the reciprocal
//  pivot growth factor are returned. This is generally more expensive
//  that LinSolve.
void SparseMatrix::LinSolveExpert(Vector &x, double &rcond, double &rpg) const
{
  if ( x.IsNull() )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::LinSolveExpert(...)'",
	    "Left vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SparseMatrix.C: In function 'SparseMatrix::LinSolveExpert(...)'",
	    "Matrix is null") ;
    }
  else if ( x.Size() != _size ) // call base-clase Size() function
    {
      error("SparseMatrix.C: In function 'SparseMatrix::LinSolveExpert(...)'",
	    "Matrix-vector size does not match") ;
    }
  else
    {
      SuperMatrix A, B, X, L, U ;
      char equed[1] ;
      void *work = 0 ;
      int lwork = 0 ; // default value
      mem_usage_t mem_usage ;
      
      // create matrix A in the format expected by SuperLU
      dCreate_CompCol_Matrix(&A,_size,_size,_nnz,_nzval,_rowind,_colptr,
			     SLU_NC,SLU_D,SLU_GE) ;

      // create right-hand side matrix B
      const int nrhs = 1 ; // number of right-hand side
      Vector b(x) ; // copy rhs vector to avoid overwrite it.
      dCreate_Dense_Matrix(&B,_size,nrhs,b.Data(),_size,SLU_DN,SLU_D,SLU_GE) ;
      dCreate_Dense_Matrix(&X,_size,nrhs,x.Data(),_size,SLU_DN,SLU_D,SLU_GE) ;

      // row permutations from partial pivoting
      int *perm_r = new int [_size] ;
      // column permutation vector
      int *perm_c = new int [_size] ;

      int *etree = new int [_size] ;
      double *R = new double [_size] ;
      double *C = new double [_size] ;
      double *ferr = new double [nrhs] ;
      double *berr = new double [nrhs] ;
 
      // set default input options
      superlu_options_t options ;
      set_default_options(&options) ;
      options.PrintStat = NO ;
      options.Equil = YES ;
      options.DiagPivotThresh = 1.0 ;
      options.PivotGrowth = YES ; // compute reciprocal pivot growth
      options.ConditionNumber = YES ; // compute reciprocal condition number.
      
      // initialize the statistics variables
      SuperLUStat_t stat ;
      StatInit(&stat) ;

      // solve the system and compute the condition number and error
      // bounds using dgssvx
      int info ;
      dgssvx(&options,&A,perm_c,perm_r,etree,equed,R,C,&L,&U,work,lwork,&B,&X,
	     &rpg,&rcond,ferr,berr,&mem_usage,&stat,&info) ;


      // checking output
      if ( info > 0 )
	{
	  if ( info <= _size )
	    {
	      cerr << "U( " << info << " , " << info << " ) is exactly zero" << endl ;
	      error("SparseMatrix.C: In function 'SparseMatrix::LinSolveExpert(...)'",
		    "Solution could not be computed, factor U is exactly singular") ;
	    }
	  else if ( info == ( _size + 1 ) )
	    {
	      cerr << endl ;
	      cerr << "\tWARNING: Matrix is singular to working precision." << endl ;
	      cerr << "\t         Check results carefully!" << endl << endl ;
	    }
	  else
	    {
	      error("SparseMatrix.C: In function 'SparseMatrix::LinSolveExpert(...)'",
		    "Memory allocation failure occured") ;
	    }
	}
      
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument had an illegal value" << endl ;
	  error("SparseMatrix.C: In function 'SparseMatrix::LinSolveExpert(...)'",
		"Illegal value in dgsvvx function!") ;
	}

      // de-allocate storage
      delete [] perm_r ;
      perm_r = 0 ;
      delete [] perm_c ;
      perm_c = 0 ;
      delete [] etree ;
      etree = 0 ;
      delete [] R ;
      R = 0 ;
      delete [] C ;
      C = 0 ;
      delete [] ferr ;
      ferr = 0 ;
      delete [] berr ;
      berr = 0 ;
      Destroy_SuperMatrix_Store(&A) ;
      Destroy_SuperMatrix_Store(&B) ;
      Destroy_SuperMatrix_Store(&X) ;
      Destroy_SuperNode_Matrix(&L) ;
      Destroy_CompCol_Matrix(&U) ;
      StatFree(&stat) ;
    }
}


//-----------------------
// This function computes the solution to a real system of linear
// equations A * x = b, using the conjugate gradient method.  where A
// is an N-by-N symmetric and positive definite sparse (Harwell-Boing
// format) matrix and x and b are N-by-1 vectors. Recommended over
// LinSolve is A is extremely large. This function implement the
// algorithm presented in:
//
// D.Kincaid and W.Cheney. Numerical Analysis: Mathematics of Scientifics
// Computing. Third Edition. Brooks/Cole Series in Advanced Mathematics.
// USA,2002. pag. 238
//
// Note: DO NOT USE if A is not symmetric and positive definite.
// ----
void SparseMatrix::ConjGradSolve(Vector &x, const int MaxIter) const
{
  if ( x.IsNull() )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::ConjGradSolve(...)'",
	    "Left vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SparseMatrix.C: In function 'SparseMatrix::ConjGradSolve(...)'",
	    "Matrix is null") ;
    }
  else if ( x.Size() != _size ) 
    {
      error("SparseMatrix.C: In function 'SparseMatrix::ConjGradSolve(...)'",
	    "Matrix-vector size does not match") ;
    }
  else
    {
      SuperMatrix A ;
      // create matrix A in the format expected by SuperLU
      dCreate_CompCol_Matrix(&A,_size,_size,_nnz,_nzval,_rowind,_colptr,
			     SLU_NC,SLU_D,SLU_GE) ;
      // conjugate gradient algorithm
      Vector r(x) ; // copy vector x into r
      // r = b - A*x
      char trans[] = "N" ;
      sp_dgemv(trans,-1.0,&A,x.Data(),1,1.0,r.Data(),1) ;
      // copy r into v
      Vector v(r) ;
      Vector z(v.Size()) ;
      double c = r.DotProduct(r) ;
      // do iteration
      for (int i=0; i<=MaxIter; i++)
	{
	  if ( i == MaxIter )
	    {
	      warning("SparseMatrix.C: In function 'SparseMatrix::ConjGradSolve(...)'",
		      "Problem does not converge after max. number of iteration, please check") ;
	      break ;
	    }
	  //cout << "tol = " << v.EuclideanNorm() << endl ;
	  if ( v.EuclideanNorm() < ZeroTol ) break ;
	  // z <- A*v
	  z.SetZero() ;
	  sp_dgemv(trans,1.0,&A,v.Data(),1,1.0,z.Data(),1) ;
	  double vz = v.DotProduct(z) ;
	  if ( vz <= ZeroTol )
	    {
	      error("SparseMatrix.C: In function 'SparseMatrix::ConjGradSolve(...)'",
		    "Division by zero!") ;
	    }
	  double t = c / vz ;
	  // x <- x + t*v
	  cblas_daxpy(v.Size(),t,v.Data(),1,x.Data(),1) ;
	  // r <- r - t*z
	  cblas_daxpy(z.Size(),-t,z.Data(),1,r.Data(),1) ;
	  double d = r.DotProduct(r) ;
	  if ( d <= ZeroTol ) break ;
	  if ( c <= ZeroTol )
	    {
	      error("SparseMatrix.C: In function 'SparseMatrix::ConjGradSolve(...)'",
		    "Division by zero!") ;
	    }
	  double dc = d / c ;
	  // v <- r + (d/c)*v
	  cblas_dscal(v.Size(),dc,v.Data(),1) ;
	  cblas_daxpy(r.Size(),1.0,r.Data(),1,v.Data(),1) ;
	  c = d ;
	}
      // destroy SuperMatrix
      Destroy_SuperMatrix_Store(&A) ;
    } 
}

//-----------------------
// this function compute the matrix-vector product and sum y' = A x + y
// where A is a sparse matrix and x and y are vectors.
// i.e. A.dgemv(x,y) ;
void SparseMatrix::dgemv(const Vector &x, Vector &y) const
{
  if ( ( x.IsNull() ) || ( y.IsNull() ) ||
       ( _status == UNINITIALIZED ) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::dgemv(...)'",
	    "Matrix/Vector uninitialized or empty") ;
    }
  else if ( y.Size() != x.Size() ) 
    {
      error("SparseMatrix.C: In function 'SparseMatrix::dgemv(...)'",
	    "Vector-Vector does not match the sizes for summation") ;
    }
  else if ( _size != x.Size() )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::dgemv(...)'",
	    "Matrix-Vector does not match the sizes for multiplication") ;
    }
  else
    {
      char trans[] = "N" ;
      // create matrix A in the format expected by SuperLU
      SuperMatrix A ;
      dCreate_CompCol_Matrix(&A,_size,_size,_nnz,_nzval,_rowind,_colptr,
			     SLU_NC,SLU_D,SLU_GE) ;
      // const_cast required to allow x to be pass to sp_dgemv as a
      // non-constant object.
      double *data = const_cast<double *>(x.Data()) ;
      sp_dgemv(trans,1.0,&A,data,1,1.0,y.Data(),1) ;
      // clean-up memory
      Destroy_SuperMatrix_Store(&A) ;
    }
}

//-----------------------
// this function compute the matrix-matrix product and sum C' = A B + C
// where A is a sparse matrix and B and C are general matrices.
// i.e. A.dgemm(B,C) ;
void SparseMatrix::dgemm(const Matrix &B, Matrix &C) const
{
  if ( ( B.GetStatus() == UNINITIALIZED ) || 
       ( C.GetStatus() == UNINITIALIZED ) ||
       ( _status == UNINITIALIZED ) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::dgemm(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else if ( ( _size != C.Rows() ) || ( B.Cols() != C.Cols() ) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::dgemm(...)'",
	    "Matrix-matrix does not match the sizes for summation") ;
    }
  else if ( _size != B.Rows() )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::dgemm(...)'",
	    "Matrix-matrix does not match the sizes for multiplication") ;
    }
  else
    {
      char trans[] = "N" ;
      // create matrix A in the format expected by SuperLU
      SuperMatrix A ;
      dCreate_CompCol_Matrix(&A,_size,_size,_nnz,_nzval,_rowind,_colptr,
			     SLU_NC,SLU_D,SLU_GE) ;
      // const_cast required to allow B to be pass to sp_dgemm as a
      // non-constant object.
      double *data = const_cast<double *>(B.Data()) ;
      sp_dgemm(trans,trans,_size,B.Cols(),_size,1.0,&A,data,B.Cols(),
	       1.0,C.Data(),_size) ;
      // clean-up memory
      Destroy_SuperMatrix_Store(&A) ;
    }
}

//-----------------------
// this function return the status of the matrix; encoding using the
// variable Status defined above
Status SparseMatrix::GetStatus() const
{
  return _status ;
}

//-----------------------

// This function clear a matrix, that is transforming it to a non
// initialized matrix. It has the same effect that the destructor. 
void SparseMatrix::Clear()
{
  if ( _nzval != 0 )
    {
      delete [] _nzval ;
      _nzval = 0 ;
    }
  if ( _rowind != 0 )
    {
      delete [] _rowind ;
      _rowind = 0 ;
    }
  if ( _colptr != 0 )
    {
      delete [] _colptr ;
      _colptr = 0 ;
    }
  _size = 0 ;      // set number of rows equal to zero
  _nnz = 0 ;       // set number of nonzeros elements in the matrix
  _status = UNINITIALIZED ;
}

//-----------------------
// this function set the size in matrix A.
// The matrix must be uninitialized.
// i.e. SparseMatrix A ; A.SetSize(4,2) ;
void SparseMatrix::SetSize(int size, int nnz)
{
  if ( _status != UNINITIALIZED )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetSize()'",
	    "Matrix was initialized!") ;
    }
  else if ( ( size <= 0 ) || ( nnz <= 0 ) )
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetSize()'",
	    "Matrix length must be a positive integer") ;
    }
  else
    {
      _nnz = nnz ;
      _status = NORMAL ; 
      
      // initialize arrays
      _nzval = new double [_nnz] ;
      _rowind = new int [_nnz] ;
      _colptr = new int [_size+1] ;
    }
}

//-----------------------
// set all the elements of sparse matrix m to the value x
// m(i,j) = x
void SparseMatrix::SetAll( double x )
{
  if ( _status != UNINITIALIZED )
    {
      for (int i=0; i<_nnz; i++) _nzval[i] = x ;
    }
  else
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetAll(...)'",
 	    "SparseMatrix uninitialized!") ;
    }
}

// multiplies the elements of sparse matrix a by the constant factor x.
// a'(i,j) = x * a(i,j)
void SparseMatrix::Scale( double x )
{
  if ( _status != UNINITIALIZED )
    {
      cblas_dscal(_nnz,x,_nzval,1) ;
    }
  else
    {
      error("SparseMatrix.C: In function 'SparseMatrix::SetAll(...)'",
 	    "SparseMatrix uninitialized!") ;
    }
}

//-----------------------

// overloded output operator for class SparseMatrix
ostream &operator<<( ostream &output, const SparseMatrix &a )
{
  if ( (a._status == UNINITIALIZED) )//|| (a._status == EMPTY) )
    {
      error("SparseMatrix.C: In function '&operator<<(...)'",
	    "SparseMatrix uninitialized or empty") ;
    }

  output << a._size << " " << a._nnz << endl ;
  for ( int i = 0; i < a._nnz ; i++ )
    {
      output << a._rowind[i] << " " ;
    }
  output << endl ;
  for ( int i = 0; i < a._size+1 ; i++ )
    {
      output << a._colptr[i] << " " ;
    }
  output << endl ;
  for ( int i = 0; i < a._nnz ; i++ )
    {
      output << a._nzval[i] << " " ;
    }
  output << endl ;

  return output ; // enables cout << x << y
}

//-----------------------
// overloded input operator for class SparseMatrix ;
// inputs values for entire SparseMatrix
istream &operator>>( istream &input, SparseMatrix &a )
{
  if ( (a._status != UNINITIALIZED) )//|| (a._status == EMPTY) )
    {
      error("SparseMatrix.C: In function '&operator>>(...)'",
	    "SparseMatrix uninitialized or empty") ;
    }

  int size, nnz ;
  input >> size >> nnz ;

  if ( ( size > 0 ) && ( nnz > 0 ) )
    {
      a._size = size ;
      a._nnz = nnz ;
      a._status = NORMAL ; 
      
      // initialize arrays
      a._nzval = new double [nnz] ;
      a._rowind = new int [nnz] ;
      a._colptr = new int [size+1] ;
    }
  else
    {
      error("SparseMatrix.C: In function '&operator>>(...)'",
	    "SparseMatrix length must be a positive integer")  ;
    }
  
  for ( int i = 0; i < nnz ; i++ )
    {
      input >> a._rowind[i] ;
    }
  for ( int i = 0; i < size+1 ; i++ )
    {
      input >> a._colptr[i] ;
    }
  for ( int i = 0; i < nnz ; i++ )
    {
      input >> a._nzval[i] ;
    }
  
  return input ;  // enables cin >> x >> y
}
