/****************************************************************************
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#include <iostream>
using std::cerr ;
using std::cout ;
using std::cin ;
using std::endl ;
#include <cstring>
#include <cmath>

#include <algorithm>
using std::min ;

#include "Vector.H"  // Vector class definition
#include "Matrix.H"
#include "SquareMatrix.H"  // Matrix class definition

//-----------------------
// default constructor for class SquareMatrix
SquareMatrix::SquareMatrix()
  // explicitly call base-class constructor
  : Matrix() 
{
  
} 

//-----------------------
// Parametrized constructor for class SquareMatrix
SquareMatrix::SquareMatrix( int size )
  // explicity call base-class parameterized constructor
  : Matrix(size,size)
{
  
}

//-----------------------
// copy constructor for class SquareMatrix
// must receive a reference to prevent infinite recursion
SquareMatrix::SquareMatrix( const SquareMatrix &SquareMatrixToCopy )
// explicitly call base-class copy constructor
  : Matrix(SquareMatrixToCopy) 
{
  
}

//-----------------------
// destructor for class SquareMatrix
SquareMatrix::~SquareMatrix()
{
 
}

//-----------------------
// this function set the number of rows and columns in matrix m.
// The matrix must be uninitialized.
// i.e. SquareMatrix m ; m.SetSize(2,2) ;
void SquareMatrix::SetSize(int size)
{
  if ( GetStatus() != UNINITIALIZED )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::SetSize()'",
	    "SquareMatrix was initialized!") ;
    }
  else if ( size <= 0  )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::SetSize()'",
	    "SquareMatrix length must be a positive integer") ;
    }
  else
    {
      Matrix::SetSize(size,size) ; // called SetSize() base-class function
    }
}

//-----------------------
// this function sets the elements of the matrix m to the corresponding
// elements of the identity matrix, m(i,j) = \delta(i,j)
void SquareMatrix::SetIdentity()
{
  if ( GetStatus() == UNINITIALIZED )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::SetIdentity()'",
	    "Matrix uninitialized!") ;
    }
  double *data = Data() ;
  const int rows = Rows() ;
  const int cols = Cols() ;
  const int N = rows * cols ;
  memset(data,0,N*sizeof(double)) ; // set all elements to zero
  for (int i=0; i<rows; i++) data[(rows+1)*i] = one ;
  SetStatus(NORMAL) ;
}

//-----------------------
// this function factorize the square matrix A into the LU decomposition
// PA = LU.
void SquareMatrix::LU_Factor()
{
  if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LU_Factor(...)'",
	    "Matrix is null") ;
    }
  else if ( GetStatus() != NORMAL ) // it was factorized?
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LU_Factor(...)'",
	    "Matrix is not in normal state! was it factorized?") ;
    }
  else
    {
      int M = Rows() ; // Number of rows of matrix A
      int N = Cols() ; // Number of columns of matrix A
      int *ipiv = new int[min(M,N)]; 
      int lda = M ; // leading dimension of array A.
      int info ;    // store error code.

      // this violate encapsulation. But, who wants to copy a 
      // big matrix twice???
      double *A = Data() ;

      // factorizing system
      info = clapack_dgetrf(CblasColMajor,M,N,A,lda,ipiv) ;

      // set new status
      SetStatus(LU_DECOMP) ;

      //    checking output info
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::LU_Factor(...)'",
		"Illegal argument in 'clapack_dgetrf_' , please check") ;
	}
      else if ( info > 0 )
	{
	  cerr << "U( " << info << " , " << info << " ) is exactly zero" << endl ;
	  warning("SquareMatrix.C: In function 'SquareMatrix::LU_Factor(...)'",
		  "Factorization has been completed, but the factor U is exactly singular, and division by zero will occur if it is used to solve a system of equations.") ;
	}
      
      // freeing memory
      delete[] ipiv ;
      ipiv = 0 ;	
    }
}

//-----------------------
// This function computes the solution to a real system of linear equations
//     A * x = b, using ATLAS library.
//  where A is an N-by-N matrix and x and b are N-by-1 vectors.
void SquareMatrix::LinSolve(Vector &x)
{
  if ( x.IsNull() )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LinSolve(...)'",
	    "Left vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LinSolve(...)'",
	    "Matrix is null") ;
    }
  else if ( x.Size() != Cols() ) // call base-clase Cols() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LinSolve(...)'",
	    "Matrix-vector size does not match") ;
    }
  else
    {
      int N = Cols() ; // Order of matrix A
      int NRHS = 1 ; // number of RHS
      int *ipiv = new int[N]; 
      int lda = N ; // leading dimension of array A.
      int ldb = N ; // leading dimension of vector x.
      int info ;    // store error code.

      // this violate encapsulation. But, who wants to copy a 
      // big matrix twice???
      double *A = Data() ;
      double *b = x.Data() ;

      // solving system
      info = clapack_dgesv(CblasColMajor,N,NRHS,A,lda,ipiv,b,ldb) ;
      
      // set new status
      SetStatus(LU_DECOMP) ;

      //    checking output info
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::LinSolve(...)'",
		"Illegal argument in 'dspsv_' , please check") ;
	}
      else if ( info > 0 )
	{
	  cerr << "D( " << info << " , " << info << " ) is exactly zero" << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::LinSolve(...)'",
		"Factorization has been completed, but the block diagonal matrix D is exactly singular, so the solution could not be computed") ;
	}
      
      // freeing memory
      delete[] ipiv ;
      ipiv = 0 ;
    }
}

//-----------------------
//  this function solves a system of linear equations
//     A * x = b
//  with a general N-by-N matrix A using the LU factorization computed
//  by LU_Factor().
void SquareMatrix::LU_Solve(Vector &x)
{
  if ( x.IsNull() )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LU_Solve(...)'",
	    "Vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LU_Solve(...)'",
	    "Matrix is null") ;
    }
  else if ( GetStatus() != LU_DECOMP )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LU_Solve(...)'",
	    "Matrix is not in a state of LU decomposition.") ;
    }
  else if ( x.Size() != Cols() ) // call base-clase Cols() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::LU_Solve(...)'",
	    "Matrix-vector size does not match") ;
    }
  else
    {
      int N = Cols() ; // Order of matrix A
      int NRHS = 1 ; // number of RHS
      int *ipiv = new int[N]; 
      int lda = N ; // leading dimension of array A.
      int ldb = N ; // leading dimension of vector x.
      int info ;    // store error code.

      // this violate encapsulation. But, who wants to copy a 
      // big matrix twice???
      double *A = Data() ;
      double *b = x.Data() ;

      // solving system
      info = clapack_dgetrs(CblasColMajor,CblasNoTrans,N,NRHS,A,lda,
			    ipiv,b,ldb) ;

      // set new status
      SetStatus(LU_DECOMP) ;

      //    checking output info
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::LU_Solve(...)'",
		"Illegal argument in 'clapack_dgetrs' , please check") ;
	}
      
      // freeing memory
      delete[] ipiv ;
      ipiv = 0 ;
    }
}  

//-----------------------
// this function compute the inverse of a matrix A from its LU decomposition
// (LU,p), storing the results in the matrix inverse. The inverse is computed 
// by solving the system A x = b for each column of the identity matrix.
void SquareMatrix::Invert(SquareMatrix &inverse) const
{
  if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Invert(...)'",
	    "Matrix is null") ;
    }
  else if ( ! inverse.IsNull() )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Invert(...)'",
	    "Input matrix was initialized") ;
    }
  else if ( Rows() != Cols() ) // call base-clase Rows() and Cols() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Invert(...)'",
	    "Matrix to invert not a square matrix") ;
    }
  else
    {
      inverse.SetSize(Rows()) ;
      int M = Rows() ; // Number of rows of matrix A
      int N = Cols() ; // Number of columns of matrix A
      int *ipiv = new int[min(M,N)]; 
      int lda = M ; // leading dimension of array A.
      int info ;    // store error code.

      // this violate encapsulation.
      double *A = new double[M*N] ;
      // copy data
      cblas_dcopy(M*N,Data(),1,A,1) ;
      //memcpy(A,Data(),M*N*sizeof(double)) ;

      // factorizing system
      info = clapack_dgetrf(CblasColMajor,M,N,A,lda,ipiv) ;

      // set new status
      //SetStatus(LU_DECOMP) ;

      //    checking output info
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::Invert()'",
		"Illegal argument in 'clapack_dgetrf_' , please check") ;
	}
      else if ( info > 0 )
	{
	  cerr << "U( " << info << " , " << info << " ) is exactly zero" << endl ;
	  warning("SquareMatrix.C: In function 'SquareMatrix::Invert()'",
		  "Factorization has been completed, but the factor U is exactly singular, and division by zero will occur if it is used to solve a system of equations.") ;
	}
          
      info = clapack_dgetri(CblasColMajor,N,A,lda,ipiv) ;

      // copy data back
      cblas_dcopy(M*N,A,1,inverse.Data(),1) ;
      //memcpy(inverse.Data(),A,M*N*sizeof(double)) ;

      // freeing memory
      delete[] A ;
      A = 0 ;
      delete[] ipiv ;
      ipiv = 0 ;
    }
}

//-----------------------
// this function compute the determinant of a matrix A from its LU
// decomposition, LU. The determinant is computed as the product of
// the diagonal elements of U and the sign of the row permutation.
// Note: This function does not check for underflow or overflow.
double SquareMatrix::Det() const 
{
  if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Det() const'",
  	    "Matrix is null") ;
      // the program never will reach this return, but the
      // gnu compiler will complain if there is not a return.
      return bigreal ; 
    }
  else if ( Rows() != Cols() ) // call base-clase Rows() and Cols() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Det() const'",
  	    "Matrix to invert not a square matrix") ;
      // the program never will reach this return, but the
      // gnu compiler will complain if there is not a return.
      return bigreal ; 
    }
  else
    {
      int info ;
      double s ; 
      int rows = Rows() ;
      int cols = Cols() ;
      int *ipiv = new int [ rows ] ;
      double det ;
      // copying src matrix data into dest matrix data to avoid
      // destruction of matrix src.
      int N =  rows * cols ;
      double *dest = new double [ N ] ;
      cblas_dcopy(N,Data(),1,dest,1) ;
      //      memcpy(dest,Data(),N*sizeof(double)) ;
      // using LU decomp to calculate determinant
      info = clapack_dgetrf(CblasColMajor,rows,cols,
			    dest,rows,ipiv);
      //    checking output info
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::Det() const'",
		"Illegal argument in 'clapack_dgetrf_' , please check") ;
	}
      // calculate determinant from A = P*L*U
      det = one ;
      s = one ;
      for (int i=0; i<rows; i++)
	{ 
	  if ( ipiv[i] != i ) s *= -one ;
	  det *= dest[rows*i+i] ;
	}
      
      // freeing memory
      delete [] ipiv ;
      ipiv = 0 ;
      delete [] dest ;
      dest = 0 ;

      // return determinant
      return s*det ;
    }
}

//-----------------------
// this funtion solve the system A x = b in-place using the Cholesky
// decomposition of A. On input x should contain the right-hand side b,
// which is replaced by the solution on output.
// The Cholesky decomposition of A can only be carried out when all
// the eigenvalues of the matrix are positive. If the matrix is not
// symmetric positive-definite then the decomposition will fail.
void SquareMatrix::Cholesky_Solve(Vector &x)
{
  if ( x.IsNull() )
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Cholesky_Solve(...)'",
	    "Left vector is null") ;
    }
  else if ( IsNull() ) // call base-clase IsNull() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Cholesky_Solve(...)'",
	    "Matrix is null") ;
    }
  else if ( x.Size() != Cols() ) // call base-clase Cols() function
    {
      error("SquareMatrix.C: In function 'SquareMatrix::Cholesky_Solve(...)'",
	    "Matrix-vector size does not match") ;
    }
  else
    {
      int N = Cols() ; // Order of matrix A
      int NRHS = 1 ; // number of RHS
      int lda = N ; // leading dimension of array A.
      int ldb = N ; // leading dimension of vector x.
      int info ;    // store error code.

      // this violate encapsulation. But, who wants to copy a 
      // big matrix twice???
      double *A = Data() ;
      double *b = x.Data() ;
      //gsl_matrix &A = this->Data() ;
      //gsl_vector &b = x.Data() ;

      // solving system
      info = clapack_dposv(CblasRowMajor,CblasUpper,N,NRHS,A,lda,b,ldb) ;

      // set new status
      SetStatus(CHOLESKY_DECOMP) ;

      //    checking output info
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::Cholesky_Solve(...)'",
		"Illegal argument in 'clapack_dposv' , please check") ;
	}
      else if ( info > 0 )
	{
	  cerr << "The leading minor of order " << info << " of A is not positive definite" 
	       << endl ;
	  error("SquareMatrix.C: In function 'SquareMatrix::Cholesky_Solve(...)'", 
		"Factorization could not be completed, and the solution has not been computed.") ;
	}
    }
}


//-------------------------------------------
//    TENSOR CLASS FUNCTIONS DEFINITIONS
//-------------------------------------------

//-----------------------
// default destructor for class Tensor
Tensor::Tensor()
  // explicitly call base-class constructor
  : SquareMatrix() 
{
  
} 

//-----------------------
// Parametrized constructor for class Tensor
Tensor::Tensor(int size)
  // explicitly call base-class constructor
  : SquareMatrix(size) 
{
  if ( (size != 2) && (size != 3) )
    {
      error("Tensor.C: In constructor 'Tensor::Tensor(int)'",
	    "Tensor size should be 2 or 3") ;
    }
} 

//-----------------------
// copy constructor for class Tensor
// must receive a reference to prevent infinite recursion
Tensor::Tensor( const Tensor &TensorToCopy )
// explicitly call base-class copy constructor
  : SquareMatrix(TensorToCopy) 
{
  
}

//-----------------------
// destructor for class Tensor
Tensor::~Tensor()
{
 
}

//-----------------------
// this function set the number of rows and columns in tensor A.
// The tensor must be uninitialized. Acceptable size are 2 or 3.
// Initialized all the elements to zero.
// i.e. Tensor A ; A.SetSize(3) ;
void Tensor::SetSize(int size)
{
  if ( GetStatus() != UNINITIALIZED )
    {
      error("Tensor.C: In function 'Tensor::SetSize(...)'",
	    "Tensor was initialized!") ;
    }
  else if ( (size != 2) && (size != 3) )
    {
      error("Tensor.C: In function 'Tensor::SetSize(...)'",
	    "Tensor size should be 2 or 3") ;
    }
  else
    {
      SquareMatrix::SetSize(size) ; // called SetSize() base-class function
    }
}

//----------------------- 
// this function calculate the invariants of a REAL SYMMETRIC second
// order tensor A, solving first for the eigenvalues and after that,
// using the following formulas:
// 2D:
//         I1 = a1 + a2
//         I2 = a1*a2 
// 3D: 
//         I1 = a1 + a2 + a3 
//         I2 = a1*a2 + a1*a3 + a2*a3
//         I3 = a1*a2*a3 
// Note that a second order tensor is represented by a square matrix
// of size 3 x 3 in 3D space and a matrix of 2 x 2 in a 2D space. If
// you try to use this function for a matrix with a different size
// the program will abort.  The result will be unexpected if you try
// to use it with an non-symmetric tensor.
void Tensor::Invariants(Vector &invariants) const
{
  // number of rows in tensor is equal to number of columns.
  const int rows = Rows() ; 
  const int cols = Cols() ;
  if ( IsNull() ) // call base-clase IsNull() function
    {
      error("Tensor.C: In function 'Tensor::Invariants(...) const'",
	    "Tensor is null") ;
    }
  else if ( rows != invariants.Size() )
    {
      error("Tensor.C: In function 'Tensor::Invariants(...) const'",
	    "Vector size should match tensor size") ;
    }
  else
    {
      // allocate space for new matrix
      const int N = rows*cols ;
      double *A = new double [ N ] ;
      cblas_dcopy(N,Data(),1,A,1)  ; 
      //memcpy(A,Data(),N*sizeof(double)) ; //copy one matrix into another
      // eigenvalue calculation
      const char *jobz = "N" ; // compute eigenvalues only
      const char *uplo = "U" ; // Upper triangle of A is stored
      int NA = rows ;
      int lda = rows ;
      int lwork = 3*rows-1 ;
      int info ;
      double *w = new double [rows] ;
      double *work = new double [lwork] ;

      // compute eigenvalues
      dsyev_(jobz,uplo,&NA,A,&lda,w,work,&lwork,&info) ;

      // checking output
      if ( info < 0 )
	{
	  cerr << "The " << -info << " argument has an illegal value." << endl ;
	  error("Tensor.C: In function 'Tensor::Invariants()'",
		"Illegal argument in 'dsyev_' , please check!") ;
	}
      else if ( info > 0 )
	{
	  error("Tensor.C: In function 'Tensor::Invariants()'",
		"The algorithm has failed to converged.") ;
	}
      else
	{
	  double lambda1, lambda2, lambda3 ; // eigenvalues
	  if ( rows == 3 ) // 3D space?
	    {
	      lambda1 = w[0] ;
	      lambda2 = w[1] ;
	      lambda3 = w[2] ;
	      // invariants
	      invariants(0) = lambda1 + lambda2 + lambda3 ;
	      invariants(1) = lambda1*lambda2 + lambda1*lambda3 + lambda2*lambda3 ;
	      invariants(2) = lambda1*lambda2*lambda3 ;
	    }
	  else // 2D space?
	    {
	      lambda1 = w[0] ;
	      lambda2 = w[1] ;
	      // invariants
	      invariants(0) = lambda1 + lambda2 ;
	      invariants(1) = lambda1*lambda2 ;
	    }
	}
      // freeing memory
      delete [] A ;
      A = 0 ;
      delete [] w ;
      w = 0 ;
      delete [] work ;
      work = 0 ;
    }
}

//-----------------------
// this function compute the trace of a tensor. tr(A) = A_ii.
double Tensor::Trace() const
{
  const int rows = Rows() ; // rows will give the tensor size
  if ( IsNull() )
    {
      error("Tensor.C: In function 'Tensor::Trace() const'",
	    "Tensor is null") ;
    }
  else if ( (rows != 2) && (rows != 3) )
    {
      error("Tensor.C: In function 'Tensor::Trace() const'",
	    "Not a second oder tensor. Size should be 2 or 3.") ;
    }
  double sum = zero ;
  for (int i=0;i<rows;i++)
    {
      sum += (*this)(i,i) ;
    }
  return sum ;
}

//-----------------------
// this function compute the contraction of the tensor A and
// B. Using A : B = A_ij B_ij.
double Tensor::Contraction(const Tensor &B) const
{
  const int rows = Rows() ; const int cols = Cols() ;
  double sum = zero ;
  if ( B.IsNull()  &&  IsNull()  ) // is A or B null?
    {
      error("Tensor.C: In function 'Tensor::Contraction(...) const'",
	    "Tensor is null") ;
    }
  else if ( rows != B.Rows() )
    {
      error("Tensor.C: In function 'Tensor::Contraction(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (rows != 2) && (rows != 3) )
    {
      error("Tensor.C: In function 'Tensor::Contraction(...) const'",
	    "Size should be 2 or 3.") ;
    }
  Tensor C(B) ;
  C *= *this ;
  for (int row = 0; row < rows; row++ )
    for (int col = 0; col < cols; col++ )
      {
	sum += C(row,col) ;
      }
  return sum ;
}

//-----------------------
// this function compute the symmetric and skew part of a tensor.
// A = S + W. S = 1/2 * (A + A^T), W = 1/2 * (A - A^T).
// S = Symmetric part of A
// W = Skew part of A
void Tensor::SymmSkew(Tensor &S, Tensor &W) const
{
  if ( ( IsNull() || S.IsNull() ) || W.IsNull() ) // are tensors null?
    {
      error("Tensor.C: In function 'Tensor::SymmSkew(...) const'",
	    "Tensor is null") ;
    }
  else if ( ( Rows() != S.Rows() ) != W.Rows() )
    {
      error("Tensor.C: In function 'Tensor::SymmSkew(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (Rows() != 2) && (Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::SymmSkew(...) const'",
	    "Size should be 2 or 3.") ;
    }
  // symmetric part
  S = *this ;
  Matrix tmp ; // temporal matrix used to store the transpose
  S.Transpose(tmp) ; // return S^T
  S += tmp ;
  S.Scale(half) ; // return the symmetric part of the tensor.
  // skew part
  W = *this ;
  tmp.Clear() ;
  W.Transpose(tmp) ;
  W -= tmp ;
  W.Scale(half) ; // return the skew part of the tensor
}

//-----------------------
// this function decomposed a tensor A in its so-called spherical
// part and its deviatoric part. i.e.
// A_ij = alpha*delta_ij + dev A_ij
// A = S + Dev
// S = spherical part
// Dev =  devaiator part
void Tensor::SphericalDeviator(Tensor &S, Tensor &Dev) const
{
  if ( ( IsNull() || S.IsNull() ) || Dev.IsNull() ) // are tensors null?
    {
      error("Tensor.C: In function 'Tensor::SphericalDeviator(...) const'",
	    "Tensor is null") ;
    }
  else if ( ( Rows() != S.Rows() ) && ( Rows() != Dev.Rows() ) )
    {
      error("Tensor.C: In function 'Tensor::SphericalDeviator(...) const'",
	    "Sizes does not match") ;
    }
  else if ( ( Rows() != 2 ) && ( Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::ShericalDeviator(...) const'",
	    "Size should be 2 or 3.") ;
    }
  // spherical part
  double alpha ;
  alpha = third * Trace() ; // call the trace of A. this->Trace().
  S.SetIdentity() ;
  S.Scale(alpha) ; // return the spherical part
  // deviator part
  Dev = *this ;
  Dev -= S ; // return the deviator part
}

//-----------------------
// this function compute the polar decomposition of the deformation
// gradient F. That is, F = RU. This function just have senses for the
// polar decomposition of the deformation gradient.
// F = deformation gradient
// R = orthogonal rotation matrix
// U = symmetric strecth matrix
// See: K.Bathe. Finite Element Procedures. Prentice Hall. NJ, 1996.
// p. 508 - 510.
void Tensor::PolarDecomposition(Tensor &R, Tensor &U) const 
{
  if ( ( IsNull() || R.IsNull() ) || U.IsNull() ) // are tensors null?
    {
      error("Tensor.C: In function 'Tensor::PolarDecomposition(...) const'",
	    "Tensor is null") ;
    }
  else if ( ( Rows() != R.Rows() ) && ( Rows() != U.Rows() ) )
    {
      error("Tensor.C: In function 'Tensor::PolarDecomposition(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (Rows() != 2) && (Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::PolarDecomposition(...) const'",
	    "Size should be 2 or 3.") ;
    }
  const int rows = Rows() ;
  const int cols = Cols() ;
  Tensor C(rows) ; // right Green-Cauchy deformation tensor. C = F^T F
  Matrix FT ; // temporal variable used to store the transpose
  Transpose(FT) ; // this->Transpose(FT)
  C.dgemm(FT,*this) ;  // this is equivalent to C = F^T F
  // allocate space for new matrix
  double *A = new double [rows*cols] ;
  cblas_dcopy(rows*cols,C.Data(),1,A,1) ;
  //  memcpy(A,C.Data(),rows*cols*sizeof(double)) ; //copy one matrix into another
  // eigenvalue calculation
  const char *jobz = "V" ; // compute eigenvalues and eigenvectors
  const char *uplo = "U" ; // Upper triangle of A is stored
  const int N = rows ;
  const int lda = rows ;
  const int lwork = 3*rows-1 ;
  int info ;
  double *w = new double [rows] ;
  double *work = new double [lwork] ;
  // compute eigenvalues
  dsyev_(jobz,uplo,&N,A,&lda,w,work,&lwork,&info) ;

  // checking output
  if ( info < 0 )
    {
      cerr << "The " << -info << " argument has an illegal value." << endl ;
      error("Tensor.C: In function 'Tensor::PolarDecomposition(...) const'",
	    "Illegal argument in 'dsyev_' , please check!") ;
    }
  else if ( info > 0 )
    {
      error("Tensor.C: In function 'Tensor::PolarDecomposition(...) const'",
	    "The algorithm has failed to converged.") ;
    }
  else
    {
      // eigenvalues of U are the sqrt of the eigenvalues of C.
      Vector lambda(rows) ;
      for (int i=0; i<rows; i++) 
	{
	  lambda(i) = sqrt( w[i] ) ;
	}
      // spectral decomposition of U
      Matrix p(rows,1) ;
      U.SetZero() ;
      for (int col=0; col<cols; col++)
	{
	  for (int row=0; row<rows; row++)
	    {
	      p(row,0) = A[rows*col+row] ;
	    }
	  Matrix pt ;
	  p.Transpose(pt) ;
	  Tensor temporal(rows) ;
	  temporal.dgemm(p,pt) ;
	  temporal.Scale( lambda(col) ) ;
	  U += temporal ; 
	}
      // orthogonal rotation matrix R = FU^-1
      R.SetZero() ;
      SquareMatrix UI ;
      U.Invert(UI) ;
      R.dgemm(*this,UI) ;
      // freeing memory
      delete [] A ;
      A = 0 ;
      delete [] w ;
      w = 0 ;
      delete [] work ;
      work = 0 ;
    }
}

//----------------------- 
// this function evaluate the transformation between the Cauchy stress
// tensor and the Second Piola-Kirchhoff stress tensor. S is the
// second piola-kirchhoff stress tensor and F is the deformation
// gradient. Both tensor should be initialized.
void Tensor::CauchyPK2(Tensor &S, const Tensor &F) const
{
  if ( S.IsNull() || IsNull()  ) // is A or B null?
    {
      error("Tensor.C: In function 'Tensor::CauchyPK2(...) const'",
	    "Tensor is null") ;
    }
  else if ( Rows() != S.Rows() )
    {
      error("Tensor.C: In function 'Tensor::CauchyPK2(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (Rows() != 2) && ( Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::CauchyPK2(...) const'",
	    "Size should be 2 or 3.") ;
    }
  // evaluation of the jacobian
  const double jac = F.Det() ;
  // S = J*F^-1 * sigma * F^-T 
  Tensor FI ;
  F.Invert(FI) ;
  Tensor FTI ;
  FI.Transpose(FTI) ;
  Tensor tmp(Rows()) ;
  tmp.dgemm(FI,*this) ;
  S.SetZero() ;
  S.dgemm(tmp,FTI) ;
  S.Scale(jac) ;
}

//----------------------- 
// this function evaluate the transformation between the Second
// Piola-Kirchhoff stress tensor and the Cauche stress tensor. sigma
// is the Cauchy stress tensor and should be initialized.
void Tensor::PK2Cauchy(Tensor &sigma, const Tensor &F) const
{
  if ( sigma.IsNull() || IsNull()  ) // is A or B null?
    {
      error("Tensor.C: In function 'Tensor::PK2Cauchy(...) const'",
	    "Tensor is null") ;
    }
  else if ( Rows() != sigma.Rows() )
    {
      error("Tensor.C: In function 'Tensor::PK2Cauchy(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (Rows() != 2) && ( Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::PK2Cauchy(...) const'",
	    "Size should be 2 or 3.") ;
    }
  const double TOL = 1.0e-12 ;
  // sigma = J^-1 * S * F^T
  // evaluation of the jacobian
  double jac_inv =  F.Det() ;
  if ( jac_inv <= TOL )
    {
      error("Tensor.C: In function 'Tensor::PK2Cauchy(...) const'",
	    "Jacobian less than or equal to zero,please check.") ;
    }
  else
    {
      jac_inv = one / jac_inv ;
    }
  // sigma = J^-1 * F * S * F^T
  Tensor FT ;
  F.Transpose(FT) ;

  Tensor tmp(F.Rows()) ; // temporal tensor to store results
  tmp.dgemm(F,*this) ; // F * S

  sigma.SetZero() ; // set all elements to zero
  sigma.dgemm(tmp,FT) ; // tmp * F^T = F * S * F^T  
  sigma.Scale(jac_inv) ;
}

//----------------------- 
// this function evaluate the transformation between the second
// Piola-Kirchhoff stress tensor and the nominal stress tensor. P is
// the nominal stress tensor and should be initialized.
void Tensor::PK2Nominal(Tensor &P, const Tensor &F) const
{
  if ( P.IsNull() || IsNull()  ) // is A or B null?
    {
      error("Tensor.C: In function 'Tensor::PK2Nominal(...) const'",
	    "Tensor is null") ;
    }
  else if ( Rows() != P.Rows() )
    {
      error("Tensor.C: In function 'Tensor::PK2Nominal(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (Rows() != 2) && ( Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::PK2Nominal(...) const'",
	    "Size should be 2 or 3.") ;
    }
  // P = S * F^T
  Tensor FT ;
  F.Transpose(FT) ;
  P.SetZero() ;
  P.dgemm(*this,FT) ;
}

//----------------------- 
// this function evaluate the transformation between the
// Green-Lagrange strain tensor and the Euler-Almansi strain tensor.
// The Euler-Almansi strain tensor should be initialized.
void Tensor::GreenAlmansi(Tensor &almansi, const Tensor &F) const
{
  if ( almansi.IsNull() || IsNull()  ) // is A or B null?
    {
      error("Tensor.C: In function 'Tensor::GreenAlmansi(...) const'",
	    "Tensor is null") ;
    }
  else if ( Rows() != almansi.Rows() )
    {
      error("Tensor.C: In function 'Tensor::GreenAlmansi(...) const'",
	    "Sizes does not match") ;
    }
  else if ( (Rows() != 2) && ( Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::GreenAlmansi(...) const'",
	    "Size should be 2 or 3.") ;
    }
  // e = F^-T * E * F^-1
  Tensor FI ;
  F.Invert(FI) ;
  Tensor FTI ;
  FI.Transpose(FTI) ;
  Tensor tmp(Rows()) ;
  tmp.dgemm(FTI,*this) ; // F^-T * E
  almansi.SetZero() ; // set all element to zero
  almansi.dgemm(tmp,FI) ; // F^-T * E * F^-1 
}

//----------------------- 
// this function transform a vector into a tensor following the
// kinetic Voigt rule.  
// See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
// Continua and Structures. Wiley. England. 2000. p.615-616.
// Note the tensor should be uninitialized.
void Tensor::KineticVoigtRule(const Vector &sigma)
{
  if ( ! IsNull() ) // is A initialized?
    {
      error("Tensor.C: In function 'Tensor::KineticVoigtRule(...)'",
	    "Tensor was initialized") ;
    }
  else if ( sigma.IsNull() ) // is sigma uninitialized?
    {
      error("Tensor.C: In function 'Tensor::KineticVoigtRule(...)'",
	    "Vector is null") ;
    }
  else if ( (sigma.Size() != 3) && (sigma.Size() != 6) )
    {
      error("Tensor.C: In function 'Tensor::KineticVoigtRule(...)'",
	    "Vector size should be 3 or 6") ;
    }
  else
    {
      const int size = sigma.Size() ;
      switch (size)
	{
	case 3:
	  SetSize(2) ; // 2 x 2 tensor in 2D.
	  (*this)(0,0) = sigma(0) ;
	  (*this)(1,1) = sigma(1) ;
	  (*this)(0,1) = sigma(2) ;
	  (*this)(1,0) = sigma(2) ;
	  break ;
	case 6:
	  SetSize(3) ; // 3 x 3 tensor in 3D.
	  (*this)(0,0) = sigma(0) ;
	  (*this)(1,1) = sigma(1) ;
	  (*this)(2,2) = sigma(2) ;
	  (*this)(1,2) = sigma(3) ;
	  (*this)(2,1) = sigma(3) ;
	  (*this)(0,2) = sigma(4) ;
	  (*this)(2,0) = sigma(4) ;
	  (*this)(0,1) = sigma(5) ;
	  (*this)(1,0) = sigma(5) ;
	  break ;
	default:
	  error("Tensor.C: In function 'Tensor::KineticVoigtRule(...)'",
		"Tensor size should be 2 or 3") ;
	  break ;
	}
    }
}

//----------------------- 
// this function transform a vector into a tensor following the
// kinematic Voigt rule.  
// See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
// Continua and Structures. Wiley. England. 2000. p.615-616.
// Note the tensor should be uninitialized.
void Tensor::KinematicVoigtRule(const Vector &epsilon)
{
  if ( ! IsNull() ) // is A initialized?
    {
      error("Tensor.C: In function 'Tensor::KinematicVoigtRule(...)'",
	    "Tensor was initialized") ;
    }
  else if ( epsilon.IsNull() ) // is epsilon uninitialized?
    {
      error("Tensor.C: In function 'Tensor::KinematicVoigtRule(...)'",
	    "Vector is null") ;
    }
  else if ( (epsilon.Size() != 3) && (epsilon.Size() != 6) )
    {
      error("Tensor.C: In function 'Tensor::KinematicVoigtRule(...)'",
	    "Vector size should be 3 or 6") ;
    }
  else
    {
      const int size = epsilon.Size() ;
      switch (size)
	{
	case 3:
	  SetSize(2) ; // 2 x 2 tensor in 2D.
	  (*this)(0,0) = epsilon(0) ;
	  (*this)(1,1) = epsilon(1) ;
	  (*this)(0,1) = half * epsilon(2) ;
	  (*this)(1,0) = half * epsilon(2) ;
	  break ;
	case 6:
	  SetSize(3) ; // 3 x 3 tensor in 3D.
	  (*this)(0,0) = epsilon(0) ;
	  (*this)(1,1) = epsilon(1) ;
	  (*this)(2,2) = epsilon(2) ;
	  (*this)(1,2) = half * epsilon(3) ;
	  (*this)(2,1) = half * epsilon(3) ;
	  (*this)(0,2) = half * epsilon(4) ;
	  (*this)(2,0) = half * epsilon(4) ;
	  (*this)(0,1) = half * epsilon(5) ;
	  (*this)(1,0) = half * epsilon(5) ;
	  break ;
	default:
	  error("Tensor.C: In function 'Tensor::KinematicVoigtRule(...)'",
		"Tensor size should be 2 or 3") ;
	  break ;
	}
    }
}

//----------------------- 
// this function transform a tensor into a vector following the
// kinetic Voigt rule.  
// See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
// Continua and Structures. Wiley. England. 2000. p.615-616.
// Note the vector to pass should be uninitialized. The return vector
// will contains the corresponding matrix values in Voigt notation.
void Tensor::KineticVoigtRule_TV(Vector &sigma) const
{
  if ( IsNull() ) // is A null?
    {
      error("Tensor.C: In function 'Tensor::KineticVoigtRule_TV(...) const'",
	    "Tensor is null") ;
    }
  else if ( ! sigma.IsNull() ) // is sigma initialized?
    {
      error("Tensor.C: In function 'Tensor::KineticVoigtRule_TV(...) const'",
	    "Vector was initialized") ;
    }
  else if ( (Rows() != 2) && (Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::KineticVoigtRule_TV(...) const'",
	    "Tensor size should be 2 or 3") ;
    }
  else
    {
      const int size = Rows() ;
      switch (size)
	{
	case 2:
	  sigma.SetSize(3) ; // we need to store three elements
	  sigma(0) = (*this)(0,0) ;
	  sigma(1) = (*this)(1,1) ;
	  sigma(2) = (*this)(0,1) ;
	  break ;
	case 3:
	  sigma.SetSize(6) ; // we need to store six elements
	  sigma(0) = (*this)(0,0) ;
	  sigma(1) = (*this)(1,1) ;
	  sigma(2) = (*this)(2,2) ;
	  sigma(3) = (*this)(1,2) ;
	  sigma(4) = (*this)(0,2) ;
	  sigma(5) = (*this)(0,1) ;
	  break ;
	default:
	  error("Tensor.C: In function 'Tensor::KineticVoigtRule_TV(...) const'",
		"Tensor size should be 2 or 3") ;
	  break ;
	}
    }
}

//----------------------- 
// this function transform a tensor into a vector following the
// kinematic Voigt rule.  
// See: T.Belytschko, W.K.Liu,B.Moran. Nonlinear Finite Elements for
// Continua and Structures. Wiley. England. 2000. p.615-616.
// Note the vector to pass should be uninitialized.
void Tensor::KinematicVoigtRule_TV(Vector &epsilon) const
{
  if ( IsNull() ) // is A null?
    {
      error("Tensor.C: In function 'Tensor::KinematicVoigtRule_TV(...) const'",
	    "Tensor is null") ;
    }
  else if ( ! epsilon.IsNull() ) // is epsilon initialized?
    {
      error("Tensor.C: In function 'Tensor::KinematicVoigtRule_TV(...) const'",
	    "Vector was initialized") ;
    }
  else if ( (Rows() != 2) && (Rows() != 3) )
    {
      error("Tensor.C: In function 'Tensor::KinematicVoigtRule_TV(...) const'",
	    "Tensor size should be 2 or 3") ;
    }
  else
    {
      const int size = Rows() ;
      switch (size)
	{
	case 2:
	  epsilon.SetSize(3) ; // we need to store three elements
	  epsilon(0) = (*this)(0,0) ;
	  epsilon(1) = (*this)(1,1) ;
	  epsilon(2) = two * ( (*this)(0,1) ) ;
	  break ;
	case 3:
	  epsilon.SetSize(6) ; // we need to store six elements
	  epsilon(0) = (*this)(0,0) ;
	  epsilon(1) = (*this)(1,1) ;
	  epsilon(2) = (*this)(2,2) ;
	  epsilon(3) = two * ( (*this)(1,2) ) ;
	  epsilon(4) = two * ( (*this)(0,2) ) ;
	  epsilon(5) = two * ( (*this)(0,1) ) ;
	  break ;
	default:
	  error("Tensor.C: In function 'Tensor::KinematicVoigtRule_TV(...) const'",
		"Tensor size should be 2 or 3") ;
	  break ;
	}
    }
}
