/****************************************************************************
 * Copyright (C) 2009 Mario J. Juha and Dr. Daniel Simkins                  *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 05/30/10    Mario Juha     Original code                                 *
 * 11/26/10    Mario Juha   Adeed function Clear()                          *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/

#include <iostream>
using std::cerr ;
using std::cout ;
using std::cin ;
using std::endl ;
#include <cstring>

#include <algorithm>
using std::min ;


#include "Vector.H"  // Vector class definition
#include "Matrix.H"

//-----------------------
// default destructor for class Vector
Vector::Vector()
{
  _data = 0 ;      // point to null
  _size =  0 ;     // set size to zero
} 

//-----------------------
// Parametrized constructor for class Vector
Vector::Vector( int n )
{
  if ( n > 0 )
    {
      // allocate space for vector and initialitize all components to zero
      _data = new double [n] ;
      _size = n ;  // set size of vector
      memset(_data,0,_size*sizeof(double)) ; // set all element to zero  
    }
  else
    {
      error("Vector.C: In function 'Vector::Vector(int)'",
	    "Vector length n must be a positive integer")  ;
    }
}

//-----------------------
// copy constructor for class Vector
// must receive a reference to prevent infinite recursion
Vector::Vector( const Vector &VectorToCopy )
{
  const double *src = VectorToCopy._data ;
  _size = VectorToCopy._size ;  // get vector size
 if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::Vector(Vector)'",
  	    "Vector length n must be a positive integer") ;
    }
  _data = new double [_size] ;  // allocate space for new vector
  cblas_dcopy(_size, src, 1, _data, 1) ; // copy data
}

//-----------------------
// destructor for class Vector
Vector::~Vector()
{
  if(_data != 0)
    {
      delete [] _data ;
      _data = 0 ;
    }
  _size = 0 ;
}

//-----------------------

bool Vector::IsNull() const
{
  return ( _size == 0 ) ;
}

//-----------------------

int Vector::Size() const
{
  return _size ;  // number of elements in array;
}

//-----------------------
// this function set the size of a previously declared vector.
// set all the elements to zero.
void Vector::SetSize(int size)
{
  if ( _size != 0 )
    { 
      error("Vector.C: In function 'Vector::SetSize(...)'",
	    "Vectors was initialized") ;
    }
  else if ( size > 0)
    {
      _size = size ;
      _data = new double [_size] ;
      memset(_data,0,_size*sizeof(double)) ; // set all element to zero
    }
  else
    {
      error("Vector.C: In function 'Vector::SetSize(...)'",
	    "Vectors size must be a positive integer") ;
    }
}

//-----------------------

// This function clear a vector, that is transforming it to a non
// initialized vector. It has the same effect that the destructor. 
void Vector::Clear()
{
  if(_data != 0)
    {
      delete [] _data ;
      _data = 0 ;
    }
  _size = 0 ;
}

//-----------------------
// this function return a pointer to the vector data.
// It is mainly intended to be used inside SquareMatrix class.
// Note that this function could be using like a modifiable lvalue and
// that if is use in a wrong way can cause problems 
double *Vector::Data()
{
  return _data ;
}  

//-----------------------
// this function return a const pointer to the vector data.
// It is mainly intended to be used inside SquareMatrix class.
// this is the const overloaded function of Data() that has been
// declared protected
const double *Vector::Data() const
{
  return _data ;
} 

//-----------------------
// overloaded assigment operator ;
// const return avoids: (a1 = a2) = a3
const Vector &Vector::operator=( const Vector &right )
{
  if ( &right != this ) // avoid self-assigment
    {
      if (  _size != right._size )
	{
	  error("Vector.C: In function '&Vector::operator='",
		"Vectors has different sizes") ;
	}
      else if ( _size == 0 )
	{
	  error("Vector.C: In function '&Vector::operator='",
		"Vector length n must be a positive integer") ;
	}
      else
	{
	  const double *src = right._data ;
	  cblas_dcopy(_size,src,1,_data,1) ;
	}
    }
  return *this ;  // enables x = y = z, for example
}

//-----------------------
// overloded subscript operator for non-const Vector;
// reference return creates a modifiable lvalue
double &Vector::operator()( int subscript )
{
  if ( ( subscript < 0 ) || ( subscript >= _size ) )
    {
      error("Vector.C: In function '&Vector::operator()'",
	    "Index out of range") ;     
    }
  return _data[subscript] ;  // reference return
}

//-----------------------
// overloded subscript operator for const Vector;
// const reference return creates an rvalue
double Vector::operator()(int subscript ) const
{
  if ( ( subscript < 0 ) || ( subscript >= _size ) )
    {
      error("Vector.C: In function 'Vector::operator()'",
	    "Index out of range") ;
    }
  return _data[subscript] ; // returns copy of this element
}

//-----------------------
// sets all the elements to the value of x
void Vector::SetAll( double x )
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::SetAll(...)'",
	    "Vector uninitialized!") ;
    }
  for (int i=0; i<_size; i++) _data[i] = x ;
}

//-----------------------
// sets all the elements to the value zero
void Vector::SetZero()
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::SetZero()'",
	    "Vector uninitialized!") ;
    }
  memset(_data,0,_size*sizeof(double)) ;
}

//-----------------------
// makes a basis vector by setting all the elements of the vector
// equal to zero except for the i-th element, which is set to one
void Vector::SetBasis( int i )
{
  if ( (_size == 0) )
    {
      error("Vector.C: In function 'Vector::SetBasis(...)'",
	    "Vector uninitialized!") ;
    }
  else if ( (i < 0) || ( i >= _size ) )
    {
      error("Vector.C: In function 'Vector::SetBasis(...)'",
	    "Index out of range!") ;
    }
  else 
    { 
      memset(_data,0,_size*sizeof(double)) ;
      _data[i] = one ;
    }
}

//-----------------------
// multiplies the elements of vector a by the constant factor x.
// a'_i = x * a_i
void Vector::Scale( double x )
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::Scale(...)'",
	    "Vector uninitialized!") ;
    }
  cblas_dscal(_size, x, _data, 1) ;
}

//-----------------------
// += overloded operator for non-const objects.
// a = a + b
Vector &Vector::operator+=( const Vector &right )
{
  if (  _size != right._size )
    {
      error("Vector.C: In function '&Vector::operator+='",
	    "Vectors has different sizes") ;
    }
  else if ( _size == 0 )
    {
      error("Vector.C: In function '&Vector::operator+='",
	    "Vector length n must be a positive integer") ;
    }
  else
    {
      cblas_daxpy(_size,one,right._data,1,_data,1) ; // a'_i = a_i + b_i
    }
  return *this ;  // return a reference to this element
}

//-----------------------
// -= overloded operator for non-const objects.
// a'_i = a_i - b_i
Vector &Vector::operator-=( const Vector &right ) 
{
  if (  _size != right._size )
    {
      error("Vector.C: In function '&Vector::operator-='",
	    "Vectors has different sizes") ;
    }
  else if ( _size == 0 )
    {
      error("Vector.C: In function '&Vector::operator-='",
	    "Vector length n must be a positive integer") ;
    }
  else
    {
      cblas_daxpy(_size,-one,right._data,1,_data,1) ; // a'_i = a_i - b_i 
    }
  return *this ;
}

//-----------------------
// *= overloded operator for non-const objects.
// a'_i = a_i * b_i
Vector &Vector::operator*=( const Vector &right )
{
  if (  _size != right._size )
    {
      error("Vector.C: In function '&Vector::operator*='",
	    "Vectors has different sizes") ;
    }
  else if ( _size == 0 )
    {
      error("Vector.C: In function '&Vector::operator*='",
	    "Vector length n must be a positive integer") ;
    }
  else
    {
      for (int i=0; i<_size; i++) 
	_data[i] = _data[i] * right._data[i] ; // a'_i = a_i * b_i
    }
  return *this ;
}

//-----------------------
// this function returns the maximum value in the vector.
double Vector::MaxValue() const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::MaxValue()'",
	    "Vector uninitialized!") ;
    }
  int index = cblas_idamax(_size,_data,1) ;
  return _data[index] ;
}

//-----------------------
// this function returns the minimum value in the vector.
double Vector::MinValue() const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::MinValue()'",
	    "Vector uninitialized!") ;
    }
  double min_val = bigreal ;
  for (int i=0; i<_size; i++)
    {
      min_val = min(_data[i],min_val) ;
    }
  return min_val ;
}

//-----------------------
// this function returns the minimim and the maximum values in
// the vector v, storing in min_out and max_out.
void Vector::MinMaxValue( double &min_out, double &max_out) const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::MinMaxValue'",
	    "Vector uninitialized!") ;
    }
  // min value
  double min_val = bigreal ;
  for (int i=0; i<_size; i++)
    {
      min_val = min(_data[i],min_val) ;
    }
  int index = cblas_idamax(_size,_data,1) ;
  // return values
  min_out = min_val ;
  max_out = _data[index] ;
}

//-----------------------
// this function returns the index of the maximum value in the vector v.
// when there are several maximum elements then the lowest index is
// returned
int Vector::MaxIndex() const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::MaxIndex()'",
	    "Vector uninitialized!") ;
    }
  return cblas_idamax(_size,_data,1) ;
}

//-----------------------
// this function returns the index of the minimum value in the vector v.
// when there are several minimum elements then the lowest index is
// returned
int Vector::MinIndex() const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::MinIndex()'",
	    "Vector uninitialized!") ;
    }
  
  // Took and modified from GNU gsl
  /* finds the smallest element of a vector */
  double min_val = bigreal ;
  int imin = 0;
  for (int i = 0; i < _size; i++)
    {
      double x = _data[i] ;
      if (x < min_val)
        {
          min_val = x ;
          imin = i ;
        }
    }
  return imin;
}

//-----------------------
// this function returns the index of the maximum and minimum values
// in the vector v, storing them in imin and imax. when there are several 
// maximum or minimum elements then the lowest indeces are returned
void Vector::MinMaxIndex( int &imin, int &imax ) const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::MinMaxIndex(...)'",
	    "Vector uninitialized!") ;
    }
  imin = MinIndex() ;
  imax = MaxIndex() ;
}

// This function compute the Euclidean norm ||x||_2 = \sqrt{\sum x_i^2}
double Vector::EuclideanNorm() const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::EuclidianNorm()'",
	    "Vector uninitialized!") ;
    }
  return cblas_dnrm2(_size,_data,1) ;
}

//-----------------------
// This function compute the absolute sum. \sum |x_i| of the elements
// of the vector x
double Vector::AbsoluteSum() const
{
  if ( _size == 0 )
    {
      error("Vector.C: In function 'Vector::AbsoluteSum()'",
	    "Vector uninitialized!") ;
    }
  return cblas_dasum(_size,_data,1) ;
}

//-----------------------
// this function compute the scalar product x^T y for the vector x and y
// i.e. x.DotProduct(y) = x^T y
double Vector::DotProduct( const Vector &y ) const
{
  if ( ( y._size == 0 ) || ( _size == 0 ) )
    {
      error("Vector.C: In function 'Vector::DotProduct()'",
	    "Vector uninitialized!") ;
    }
  if ( y._size != _size )
    {
      error("Vector.C: In function 'Vector::DotProduct()'",
	    "Vectors of different sizes!") ;
    }
  return cblas_ddot(_size, _data, 1, y._data, 1) ;
} 

//-----------------------
// this function compute the matrix-vector product and sum y' = A x + y
// i.e. Vector x(2) ; x.SetAll(1.0) ; Matrix A(2,2) ;
//      A.setAll(2.0) ; Vector y(2) ; y.SetAll(3.0) ; y.dgemv(A,x) ; 
void Vector::dgemv(const Matrix &A , const Vector &x)
{
  if ( ( _size == 0 ) || ( x._size == 0 ) )
    {
      error("Vector.C: In function 'Vector::dgemv(...)'",
	    "Vector uninitialized or empty") ;
    }
  else if ( A.IsNull() )
    {
      error("Vector.C: In function 'Vector::dgemv(...)'",
	    "Matrix uninitialized or empty") ;
    }
  else if ( x._size != A.Cols() )
    {
      error("Vector.C: In function 'Vector::dgemv(...)'",
	    "Matrix-vector does not match sizes for multiplication") ;
    }
  else
    {
      cblas_dgemv(CblasColMajor,CblasNoTrans,A.Rows(),A.Cols(),
		  one,A.Data(),A.Rows(),x._data,1,one,_data,1) ;
    }
}

//-----------------------
// overloded input operator for class Vector ;
// inputs values for entire Vector
istream &operator>>( istream &input, Vector &a )
{
  if ( a._size == 0 )
    {
      error("Vector.C: In function 'istream &operator>>(...)'",
	    "Vector uninitialized!") ;
    }
  for ( int i = 0; i < a._size; i++ )
    input >> a(i) ;

  return input ;  // enables cin >> x >> y
}

//-----------------------
// overloded output operator for class Vector
ostream &operator<<( ostream &output, const Vector &a )
{
  if ( a._size == 0 )
    {
      error("Vector.C: In function 'istream &operator<<(...)'",
	    "Vector uninitialized!") ;
    }
  int i ;
  // output private ptr-based array
  for ( i = 0; i < a._size; i++ )
    output << a(i) << ' ' ;

  return output ; // enables cout << x << y
}

//-----------------------

