/****************************************************************************
 * Test program for some functions in the library. This is intended to be   *
 * an example of how to use it.                                             *
 * Copyright (C) 2009 Mario J. Juha                                         *
 * Written by Mario J. Juha <mariojesusjuha@gmail.com>                      *
 *                                                                          * 
 *           University of South Florida                                    *
 *        Civil and Environmental Engineering                               *
 *                 Tampa - Florida                                          *
 * Record of revisions:                                                     *
 *   Date      Programmer   Description of change                           *
 * ========    ==========   =====================                           *
 * 07/20/09    Mario Juha     Original code                                 *
 *                                                                          *
 * Description:                                                             *  
 * ------------                                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 ***************************************************************************/
#include <iostream>
using std::cout ;
using std::endl ;

#include <fstream>
using std::ifstream ;

#include <cmath>

#include "Const_num.H"
#include "Vector.H"
#include "Matrix.H"
#include "SquareMatrix.H"
#include "SparseMatrix.H"

int main()
{
  // ------------------------------------
  //     TEST VECTOR AND MATRIX CLASS
  // ------------------------------------

  // This part shows how to allocate and initialize a vector
  // this create a vector of size 5 and set all components to zero.
  Vector v(5) ;
  // filling the vector
  for (int i = 0; i < 5; i++)
    {
      v(i) = 1.23 + i ;
    }
  // example of how to print the vector
  cout << endl << "The vector 'v' is:" << endl ;
  cout << v << endl << endl ;
  // example of how to extract the size of the vector
  cout << "The size of 'v' is " << v.Size() << endl << endl ;
  // example of how to declare a vector and initialize it
  Vector vv ;
  vv.SetSize(8) ; // all elements are now zero
  vv.SetAll(3.0) ; // all elements set to 3.0
  cout  << "The vector 'vv' is:" << endl ;
  cout << vv << endl << endl ;
  // example of how to extract max element
  cout << "Maximun value in vector 'v' is "<< v.MaxValue() 
       << endl << endl ;
  // example of how to calculate the euclidian norm
  cout << "Euclidian norm of vector 'vv' is " << vv.EuclideanNorm()
       << endl << endl ;
  // example of how to copy a vector and take the dot product
  // of two vectors
  Vector vvv(vv) ;
  cout << "The dot product of 'vv' to itself is " << vv.DotProduct(vvv)
       << endl << endl ;

  // This part shows how to allocate and initialize a matrix
  // this create a matrix of size 3 x 4 and set all components to zero.
  Matrix A(3,5) ;
  // filling the matrix and using functions Rows() and Cols()
  for (int i = 0; i < A.Rows(); i++)
    for (int j = 0; j < A.Cols(); j++)
      A(i,j) = 0.23 + 100*i + j ;
 
  // example of how to print the matrix
  cout << "The matrix 'A' is:" << endl ;
  cout << A << endl ;

  // example of how multiply a vector and matrix
  Vector y(3) ;
  y.dgemv(A,v) ;
  cout << "The product of A*v is " << endl << y 
       << endl << endl ;
  // example of how to solve a linear system
  SquareMatrix K(4) ;
  K(0,0) = 0.18; K(0,1) = 0.60; K(0,2) = 0.57; K(0,3) = 0.96;
  K(1,0) = 0.41; K(1,1) = 0.24; K(1,2) = 0.99; K(1,3) = 0.58;
  K(2,0) = 0.14; K(2,1) = 0.30; K(2,2) = 0.97; K(2,3) = 0.66;
  K(3,0) = 0.51; K(3,1) = 0.13; K(3,2) = 0.19; K(3,3) = 0.85;

  Vector b(4) ;
  for (int i=0; i<b.Size(); i++)
    b(i) = i + one ;

  cout << "The matrix 'K' is:" << endl ;
  cout << K << endl ;
  cout << "The vector 'b' is:" << endl ;
  cout << b << endl << endl ;

  // Determinant of K
  cout << "The determinant of K is " << K.Det() << endl << endl ;

  // example of how to extract the determinant of a matrix
  //cout << "Determinant of K is " << K.Det() << endl << endl ;

  cout << "The solution of u = K^{-1}b is " << endl ;
  SquareMatrix KInv ;
  K.Invert(KInv) ;
  Vector u( K.Rows() ) ;
  u.dgemv(KInv,b) ;
  cout << u << endl << endl ;

  K.LinSolve(b) ;

  cout << "The solution of Ku = b using LU decomposition is " << endl ;
  cout << b << endl << endl ;

  // using Cholesky decomposition
  SquareMatrix Q(4) ;
  Q(0,0) = 136.01 ;  Q(0,1) = 90.860 ;
  Q(1,0) = 90.860 ;  Q(1,1) = 98.810 ; Q(1,2) = -67.590 ;
  Q(2,1) = -67.590 ; Q(2,2) = 132.01 ; Q(2,3) = 46.260 ;
  Q(3,2) = 46.260 ;  Q(3,3) = 177.17 ;

  cout << "The matrix 'Q' is:" << endl ;
  cout << Q << endl ;
  cout << "The vector 'b' is:" << endl ;
  cout << b << endl << endl ;

  cout << "The solution of u = Q^{-1}b is " << endl ;
  SquareMatrix QInv ;
  Q.Invert(QInv) ;
  u.SetZero() ;
  u.dgemv(QInv,b) ;
  cout << u << endl << endl ;

  Q.Cholesky_Solve(b) ;

  cout << "The solution of Qu = b using Cholesky decomposition is " << endl ;
  cout << b << endl << endl ;

  // ------------------------------------
  //        TEST SPARSEMATRIX CLASS
  // ------------------------------------

  cout << " ==========================================" ;
  cout << endl << "      Testing sparse matrix class" << endl ;
  cout << " ==========================================" ;
  cout << endl ;


  const int msize = 5 ; // matrix size
  const int nnz = 12 ;  // number of nonzero elements

  SparseMatrix SMA(msize,nnz) ;
  // get row index from SMA
  int *asub = SMA.GetRowIndexArray() ;
  // get column pointer array
  int *xa = SMA.GetColIndexArray() ;

  // fill in row index and column pointer (sparcity)
  asub[0] = 0 ; asub[1] = 1 ; asub[2] = 4 ; asub[3] = 1 ;
  asub[4] = 2 ; asub[5] = 4 ; asub[6] = 0 ; asub[7] = 2 ;
  asub[8] = 0 ; asub[9] = 3 ; asub[10] = 3 ; asub[11] = 4 ;
  xa[0] = 0 ; xa[1] = 3 ; xa[2] = 6 ; xa[3] = 8 ; xa[4] = 10 ; xa[5] = 12 ;

  SMA(0,0) = 19.0 ; SMA(0,2) = 21.0 ; SMA(0,3) = 21.0 ;
  SMA(1,0) = 12.0 ; SMA(1,1) = 21.0 ; SMA(2,1) = 12.0 ;
  SMA(2,2) = 16.0 ; SMA(3,3) = 5.0 ; SMA(3,4) = 21.0 ;
  SMA(4,0) = 12.0 ; SMA(4,1) = 12.0 ; SMA(4,4) = 18.0 ;

  Vector x(SMA.Size()) ;
  x.SetAll(one) ;
  
  double rcond, rpg ;
  SMA.LinSolveExpert(x,rcond,rpg) ;
  //SMA.LinSolve(x) ;

  cout << "The solution is " << endl ;
  cout << x << endl ;

  cout << endl << "rcond = " << rcond << endl ;

  cout << endl ;


  // ------------------------------------
  //          TEST TENSOR CLASS
  // ------------------------------------

  // Example of how to use the tensor library.
  Tensor F(2) ; // deformation gradient.
  F(0,0) = half ; F(0,1) = -fourth ;
  F(1,0) = one ; F(1,1) = half ; 

  cout << "Deformation gradient is " << endl << F << endl ;

  // Invariants of the right Green-Cauchy deformation tensor
  Tensor FT ;
  F.Transpose(FT) ;
  Tensor C(2) ;  // right Green-Cauchy deformation tensor
  C.dgemm(FT,F) ;

  cout << "right Green-Cauchy deformation tensor is " << endl << C << endl ;

  Vector invar(2) ;
  double i1,i2 ;  // invariants
  C.Invariants(invar) ;

  i1 = invar(0) ;
  i2 = invar(1) ;

  cout << "invariants of the right Green-Cauchy deformation tensor are" << endl ;
  cout << "i1 = " << i1 << endl ;
  cout << "i2 = " << i2 << endl ;

  // spherical and deviator part of the deformation gradient
  Tensor S(2) ;
  Tensor Dev(2) ;
  F.SphericalDeviator(S,Dev) ;

  cout << endl ;
  cout << "Spherical part of the deformation gradient is" << endl << S << endl ;
  cout << "Deviatoric part of the deformation gradient is " << endl << Dev << endl ;

  // polar decomposition of the deformation gradiant.
  Tensor R(2) ; // orthogonal rotation matrix
  Tensor U(2) ; // symmetric stretch matrix
  F.PolarDecomposition(R,U) ;

  cout << endl ;
  cout << "Polar decomposition of the deformation gradient" << endl << endl ;
  cout << "Orthogonal rotation matrix is" << endl << R << endl ;
  cout << "Symmetric stretch matrix is" << endl << U << endl ;

  // kinetic Voigt rule
  Tensor sigma ;
  Vector v_sigma(6) ;
  for (int i=0;i<6;i++) v_sigma(i) = double(i) ;
  sigma.KineticVoigtRule(v_sigma) ;
  cout << "Testing kinetic Voigt rule:" << endl << endl ;
  cout << "vector sigma" << endl << v_sigma << endl << endl ;
  cout << "matrix sigma" << endl << sigma << endl ;
  cout << endl ;
  
  // kinematic Voigt rule
  Tensor epsilon ;
  Vector v_epsilon(6) ;
  for (int i=0;i<6;i++) v_epsilon(i) = double(i) ;
  epsilon.KinematicVoigtRule(v_epsilon) ;
  cout << "Testing kinematic Voigt rule:" << endl << endl ;
  cout << "vector epsilon" << endl << v_epsilon << endl << endl ;
  cout << "matrix epsilon" << endl << epsilon << endl ;
  cout << endl ;

  return (0) ;
}



